=== added file 'DBUS-API'
--- DBUS-API	1970-01-01 00:00:00 +0000
+++ DBUS-API	2010-09-12 03:00:40 +0000
@@ -0,0 +1,147 @@
+		   -*- mode: org; coding: utf-8 -*-
+
+		    Mandos Server D-Bus Interface
+
+This file documents the D-Bus interface to the Mandos server.
+
+* Bus: System bus
+  Bus name: "se.bsnet.fukt.Mandos"
+
+
+* Object Paths:
+  
+  | Path                  | Object            |
+  |-----------------------+-------------------|
+  | "/"                   | The Mandos Server |
+  | "/clients/CLIENTNAME" | Mandos Client     |
+
+  
+* Mandos Server Interface:
+  Interface name: "se.bsnet.fukt.Mandos"
+  
+** Methods:
+*** GetAllClients() → (ao: Clients)
+    Returns an array of all client D-Bus object paths
+   
+*** GetAllClientsWithProperties() → (a{oa{sv}}: ClientProperties)
+    Returns an array of all clients and all their properties
+   
+*** RemoveClient(o: ObjectPath) → nothing
+    Removes a client
+   
+** Signals:
+*** ClientAdded(o: ObjectPath)
+    A new client was added.
+   
+*** ClientNotFound(s: Fingerprint, s: Address)
+    A client connected from Address using Fingerprint, but was
+    rejected because it was not found in the server.  The fingerprint
+    is represented as a string of hexadecimal digits.  The address is
+    an IPv4 or IPv6 address in its normal string format.
+   
+*** ClientRemoved(o: ObjectPath, s: Name)
+    A client named Name on ObjectPath was removed.
+
+
+* Mandos Client Interface:
+  Interface name: "se.bsnet.fukt.Mandos.Client"
+  
+** Methods
+*** Approve(b: Approve) → nothing
+    Approve or deny a connected client waiting for approval.  If
+    denied, a client will not be sent its secret.
+    
+*** CheckedOK() → nothing
+    Assert that this client has been checked and found to be alive.
+    This will restart the timeout before disabling this client.  See
+    also the "LastCheckedOK" property.
+    
+*** Disable() → nothing
+    Disable this client.  See also the "Enabled" property.
+    
+*** Enable() → nothing
+    Enable this client.  See also the "Enabled" property.
+    
+*** StartChecker() → nothing
+    Start a new checker for this client, if none is currently
+    running.  See also the "CheckerRunning" property.
+    
+*** StopChecker() → nothing
+    Abort a running checker process for this client, if any.  See also
+    the "CheckerRunning" property.
+
+** Properties
+   
+   Note: Many of these properties directly correspond to a setting in
+   "clients.conf", in which case they are fully documented in
+   mandos-clients.conf(5).
+   
+   | Name                 | Type | Access     | clients.conf        |
+   |----------------------+------+------------+---------------------|
+   | ApprovedByDefault    | b    | Read/Write | approved_by_default |
+   | ApprovalDelay (a)    | t    | Read/Write | approval_delay      |
+   | ApprovalDuration (a) | t    | Read/Write | approval_duration   |
+   | ApprovalPending (b)  | b    | Read       | N/A                 |
+   | Checker              | s    | Read/Write | checker             |
+   | CheckerRunning (c)   | b    | Read/Write | N/A                 |
+   | Created (d)          | s    | Read       | N/A                 |
+   | Enabled (e)          | b    | Read/Write | N/A                 |
+   | Fingerprint          | s    | Read       | fingerprint         |
+   | Host                 | s    | Read/Write | host                |
+   | Interval (a)         | t    | Read/Write | interval            |
+   | LastCheckedOK (f)    | s    | Read/Write | N/A                 |
+   | LastEnabled (g)      | s    | Read       | N/A                 |
+   | Name                 | s    | Read       | (Section name)      |
+   | ObjectPath           | o    | Read       | N/A                 |
+   | Secret (h)           | ay   | Write      | secret (or secfile) |
+   | Timeout (a)          | t    | Read/Write | timeout             |
+   
+   a) Represented as milliseconds.
+   
+   b) An approval is currently pending.
+   
+   c) Setting this property is equivalent to calling StartChecker() or
+      StopChecker().
+   
+   d) The creation time of this client object, as a RFC 3339 string.
+   
+   e) Setting this property is equivalent to calling Enable() or
+      Disable().
+   
+   f) The last time a checker was successful, as a RFC 3339 string, or
+      an empty string if this has not happened.  Setting this property
+      is equivalent to calling CheckedOK(), i.e. the current time is
+      set, regardless of the string sent.  Please always use an empty
+      string when setting this property, to allow for possible future
+      expansion.
+   
+   g) The last time this client was enabled, as a RFC 3339 string, or
+      an empty string if this has not happened.
+   
+   h) A raw byte array, not hexadecimal digits.
+
+** Signals
+*** CheckerCompleted(n: Exitcode, x: Waitstatus, s: Command)
+    A checker (Command) has completed.  Exitcode is either the exit
+    code or -1 for abnormal exit.  In any case, the full Waitstatus
+    (as from wait(2)) is also available.
+    
+*** CheckerStarted(s: Command)
+    A checker command (Command) has just been started.
+    
+*** GotSecret()
+    This client has been sent its secret.
+    
+*** NeedApproval(t: Timeout, b: ApprovedByDefault)
+    This client will be approved or denied in exactly Timeout
+    milliseconds, depending on ApprovedByDefault.  Approve() can now
+    usefully be called on this client object.
+
+*** PropertyChanged(s: Property, v: Value)
+    The Property on this client has changed to Value.
+
+*** Rejected(s: Reason)
+    This client was not given its secret for a specified Reason.
+
+
+#+STARTUP: showall

=== modified file 'Makefile'
--- Makefile	2009-11-19 18:31:28 +0000
+++ Makefile	2010-09-12 03:22:48 +0000
@@ -83,7 +83,8 @@
 	--expression='s/\(<a class="citerefentry" href="\)\("><span class="citerefentry"><span class="refentrytitle">\)\([^<]*\)\(<\/span>(\)\([^)]*\)\()<\/span><\/a>\)/\1\3.\5\2\3\4\5\6/g'
 
 PLUGINS=plugins.d/password-prompt plugins.d/mandos-client \
-	plugins.d/usplash plugins.d/splashy plugins.d/askpass-fifo
+	plugins.d/usplash plugins.d/splashy plugins.d/askpass-fifo \
+	plugins.d/plymouth
 CPROGS=plugin-runner $(PLUGINS)
 PROGS=mandos mandos-keygen mandos-ctl $(CPROGS)
 DOCS=mandos.8 plugin-runner.8mandos mandos-keygen.8 \
@@ -209,6 +210,17 @@
 
 # Run the client with a local config and key
 run-client: all keydir/seckey.txt keydir/pubkey.txt
+	@echo "###################################################################"
+	@echo "# The following error messages are harmless and can be safely     #"
+	@echo "# ignored.  The messages are caused by not running as root, but   #"
+	@echo "# you should NOT run \"make run-client\" as root unless you also    #"
+	@echo "# unpacked and compiled Mandos as root, which is NOT recommended. #"
+	@echo "# From plugin-runner: setuid: Operation not permitted             #"
+	@echo "# From askpass-fifo:  mkfifo: Permission denied                   #"
+	@echo "# From mandos-client: setuid: Operation not permitted             #"
+	@echo "#                     seteuid: Operation not permitted            #"
+	@echo "#                     klogctl: Operation not permitted            #"
+	@echo "###################################################################"
 	./plugin-runner --plugin-dir=plugins.d \
 		--config-file=plugin-runner.conf \
 		--options-for=mandos-client:--seckey=keydir/seckey.txt,--pubkey=keydir/pubkey.txt \
@@ -221,6 +233,12 @@
 
 # Run the server with a local config
 run-server: confdir/mandos.conf confdir/clients.conf
+	@echo "#################################################################"
+	@echo "# NOTE: Please IGNORE the error about \"Could not open file      #"
+	@echo "# u'/var/run/mandos.pid'\" -  it is harmless and is caused by    #"
+	@echo "# the server not running as root.  Do NOT run \"make run-server\" #"
+	@echo "# server as root if you didn't also unpack and compile it thus. #"
+	@echo "#################################################################"
 	./mandos --debug --no-dbus --configdir=confdir $(SERVERARGS)
 
 # Used by run-server
@@ -288,6 +306,9 @@
 	install --mode=u=rwxs,go=rx \
 		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
 		plugins.d/askpass-fifo
+	install --mode=u=rwxs,go=rx \
+		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
+		plugins.d/plymouth
 	install initramfs-tools-hook \
 		$(INITRAMFSTOOLS)/hooks/mandos
 	install --mode=u=rw,go=r initramfs-tools-hook-conf \
@@ -338,6 +359,7 @@
 		$(PREFIX)/lib/mandos/plugins.d/usplash \
 		$(PREFIX)/lib/mandos/plugins.d/splashy \
 		$(PREFIX)/lib/mandos/plugins.d/askpass-fifo \
+		$(PREFIX)/lib/mandos/plugins.d/plymouth \
 		$(INITRAMFSTOOLS)/hooks/mandos \
 		$(INITRAMFSTOOLS)/conf-hooks.d/mandos \
 		$(INITRAMFSTOOLS)/scripts/init-premount/mandos \

=== modified file 'TODO'
--- TODO	2010-08-22 13:34:42 +0000
+++ TODO	2010-09-12 18:23:40 +0000
@@ -1,11 +1,13 @@
 -*- org -*-
 
-* _attribute_((nonnull))
+* Use _attribute_((nonnull)) wherever possible.
+
+* Release critical
+** plymouth.8mandos
 
 * mandos-client
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO use error() instead of perror()
 ** TODO [#B] Retry a server which has a non-definite reply:
 *** A closed connection during the TLS handshake
 *** A TCP timeout
@@ -15,33 +17,31 @@
 * splashy
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 
 * usplash
 ** TODO [#A] Make it work again
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
+** TODO Use [[info:libc:Argz%20Functions][argz_extract]]
 
 * askpass-fifo
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 ** TODO [#B] Drop privileges after opening FIFO.
 
 * password-prompt
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 ** TODO [#B] lock stdin (with flock()?)
 
+* plymouth
+** TODO Use [[info:libc:Argz%20Functions][argz_extract]]
+
 * TODO [#B] passdev
 
 * plugin-runner
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#C] use same file name rules as run-parts(8)
 ** kernel command line option for debug info
-** TODO [#B] use error() instead of perror()
-** TODO [$B] Use openat() and readdir64()
-   http://udrepper.livejournal.com/19395.html
+** TODO [#B] Use openat()
 
 * mandos (server)
 ** TODO [#B] Log level							  :BUGS:
@@ -75,24 +75,22 @@
 ** More D-Bus methods
 *** NeedsApproval(50, True) -> timeout, default approve
     Default approval is configurable, but True by default
-    + Approval(True) -> approve sending saved
-    + Approval(False) -> Close client connection immediately
+    + Approve(True) -> approve sending saved
+    + Approve(False) -> Close client connection immediately
 *** NeedsPassword(50) - Timeout, default disapprove
     + SetPass(u"gazonk", True) -> Approval, persistent
-    + Approval(False) -> Close client connection immediately
+    + Approve(False) -> Close client connection immediately
 ** TODO [#C] python-parsedatetime
 ** TODO [#C] systemd/launchd
    http://0pointer.de/blog/projects/systemd.html
 ** TODO Separate logging logic to own object
 ** TODO make clients to a dict!
+** TODO [#A] Limit approval_delay to max gnutls/tls timeout value
+** TODO [#B] break the wait on approval_delay if connection dies
 
 * mandos.xml
-** [[file:mandos.xml::XXX][Document D-Bus interface]]
-   Remove mention of lack of such interface in BUGS section
 ** Add mandos contact info in manual pages
 
-* TODO [#A] Provide and install /etc/dbus-1/system.d/mandos.conf
-
 * mandos-ctl
 *** Handle "no D-Bus server" and/or "no Mandos server found" better
 *** [#B] --dump option
@@ -103,9 +101,12 @@
   arguments.
 
 * mandos-monitor
+** TODO help should be toggable
 ** Urwid client data displayer
    Better view of client data in the listing
 *** Properties popup
+** Nicer crashes. Stack traces Messes up shell.
+*** Print a nice "We are sorry" message, save stack trace to log.
 
 * mandos-keygen
 ** TODO Loop until passwords match when run interactively

=== modified file 'clients.conf'
--- clients.conf	2009-12-25 23:13:47 +0000
+++ clients.conf	2010-09-12 03:00:40 +0000
@@ -16,6 +16,15 @@
 # What command to run as "the checker".
 ;checker = fping -q -- %%(host)s
 
+# Whether to approve a client by default after the approval delay.
+;approved_by_default = True
+
+# How long to wait for approval.
+;approval_delay = 0s
+
+# How long one approval will last.
+;approval_duration = 1s
+
 
 ;####
 ;# Example client
@@ -62,4 +71,9 @@
 ;
 ;# Parameters from the [DEFAULT] section can be overridden per client.
 ;interval = 5m
+;
+;# This client requires manual approval before it receives its secret.
+;approved_by_default = False
+;# Require approval within 30 seconds.
+;approval_delay = 30s
 ;####

=== modified file 'debian/mandos.docs'
--- debian/mandos.docs	2008-10-18 11:17:22 +0000
+++ debian/mandos.docs	2010-09-12 03:00:40 +0000
@@ -1,3 +1,4 @@
 NEWS
 README
 TODO
+DBUS-API

=== modified file 'mandos'
--- mandos	2010-08-24 18:18:01 +0000
+++ mandos	2010-09-12 18:23:40 +0000
@@ -83,7 +83,8 @@
 
 version = "1.0.14"
 
-logger = logging.getLogger(u'mandos')
+#logger = logging.getLogger(u'mandos')
+logger = logging.Logger(u'mandos')
 syslogger = (logging.handlers.SysLogHandler
              (facility = logging.handlers.SysLogHandler.LOG_DAEMON,
               address = "/dev/log"))
@@ -98,8 +99,6 @@
                                        u' %(message)s'))
 logger.addHandler(console)
 
-multiprocessing_manager = multiprocessing.Manager()
-
 class AvahiError(Exception):
     def __init__(self, value, *args, **kwargs):
         self.value = value
@@ -158,15 +157,20 @@
                             u" after %i retries, exiting.",
                             self.rename_count)
             raise AvahiServiceError(u"Too many renames")
-        self.name = self.server.GetAlternativeServiceName(self.name)
+        self.name = unicode(self.server.GetAlternativeServiceName(self.name))
         logger.info(u"Changing Zeroconf service name to %r ...",
-                    unicode(self.name))
+                    self.name)
         syslogger.setFormatter(logging.Formatter
                                (u'Mandos (%s) [%%(process)d]:'
                                 u' %%(levelname)s: %%(message)s'
                                 % self.name))
         self.remove()
-        self.add()
+        try:
+            self.add()
+        except dbus.exceptions.DBusException, error:
+            logger.critical(u"DBusException: %s", error)
+            self.cleanup()
+            os._exit(1)
         self.rename_count += 1
     def remove(self):
         """Derived from the Avahi example code"""
@@ -195,7 +199,7 @@
         self.group.Commit()
     def entry_group_state_changed(self, state, error):
         """Derived from the Avahi example code"""
-        logger.debug(u"Avahi state change: %i", state)
+        logger.debug(u"Avahi entry group state change: %i", state)
         
         if state == avahi.ENTRY_GROUP_ESTABLISHED:
             logger.debug(u"Zeroconf service established.")
@@ -214,6 +218,7 @@
             self.group = None
     def server_state_changed(self, state):
         """Derived from the Avahi example code"""
+        logger.debug(u"Avahi server state change: %i", state)
         if state == avahi.SERVER_COLLISION:
             logger.error(u"Zeroconf server name collision")
             self.remove()
@@ -260,9 +265,9 @@
                      runtime with vars(self) as dict, so that for
                      instance %(name)s can be used in the command.
     current_checker_command: string; current running checker_command
-    approved_delay: datetime.timedelta(); Time to wait for approval
+    approval_delay: datetime.timedelta(); Time to wait for approval
     _approved:   bool(); 'None' if not yet approved/disapproved
-    approved_duration: datetime.timedelta(); Duration of one approval
+    approval_duration: datetime.timedelta(); Duration of one approval
     """
     
     @staticmethod
@@ -280,8 +285,8 @@
         "Return the 'interval' attribute in milliseconds"
         return self._timedelta_to_milliseconds(self.interval)
 
-    def approved_delay_milliseconds(self):
-        return self._timedelta_to_milliseconds(self.approved_delay)
+    def approval_delay_milliseconds(self):
+        return self._timedelta_to_milliseconds(self.approval_delay)
     
     def __init__(self, name = None, disable_hook=None, config=None):
         """Note: the 'checker' key in 'config' sets the
@@ -305,7 +310,6 @@
                       "rb") as secfile:
                 self.secret = secfile.read()
         else:
-            #XXX Need to allow secret on demand!
             raise TypeError(u"No secret or secfile for client %s"
                             % self.name)
         self.host = config.get(u"host", u"")
@@ -323,16 +327,16 @@
         self.checker_command = config[u"checker"]
         self.current_checker_command = None
         self.last_connect = None
-        self.approvals_pending = 0
         self._approved = None
         self.approved_by_default = config.get(u"approved_by_default",
                                               True)
-        self.approved_delay = string_to_delta(
-            config[u"approved_delay"])
-        self.approved_duration = string_to_delta(
-            config[u"approved_duration"])
+        self.approvals_pending = 0
+        self.approval_delay = string_to_delta(
+            config[u"approval_delay"])
+        self.approval_duration = string_to_delta(
+            config[u"approval_duration"])
         self.changedstate = multiprocessing_manager.Condition(multiprocessing_manager.Lock())
-
+    
     def send_changedstate(self):
         self.changedstate.acquire()
         self.changedstate.notify_all()
@@ -702,6 +706,7 @@
     # dbus.service.Object doesn't use super(), so we can't either.
     
     def __init__(self, bus = None, *args, **kwargs):
+        self._approvals_pending = 0
         self.bus = bus
         Client.__init__(self, *args, **kwargs)
         # Only now, when this client is initialized, can it show up on
@@ -711,6 +716,22 @@
                                   + self.name.replace(u".", u"_")))
         DBusObjectWithProperties.__init__(self, self.bus,
                                           self.dbus_object_path)
+        
+    def _get_approvals_pending(self):
+        return self._approvals_pending
+    def _set_approvals_pending(self, value):
+        old_value = self._approvals_pending
+        self._approvals_pending = value
+        bval = bool(value)
+        if (hasattr(self, "dbus_object_path")
+            and bval is not bool(old_value)):
+            dbus_bool = dbus.Boolean(bval, variant_level=1)
+            self.PropertyChanged(dbus.String(u"ApprovalPending"),
+                                 dbus_bool)
+
+    approvals_pending = property(_get_approvals_pending,
+                                 _set_approvals_pending)
+    del _get_approvals_pending, _set_approvals_pending
     
     @staticmethod
     def _datetime_to_dbus(dt, variant_level=0):
@@ -723,10 +744,10 @@
         r = Client.enable(self)
         if oldstate != self.enabled:
             # Emit D-Bus signals
-            self.PropertyChanged(dbus.String(u"enabled"),
+            self.PropertyChanged(dbus.String(u"Enabled"),
                                  dbus.Boolean(True, variant_level=1))
             self.PropertyChanged(
-                dbus.String(u"last_enabled"),
+                dbus.String(u"LastEnabled"),
                 self._datetime_to_dbus(self.last_enabled,
                                        variant_level=1))
         return r
@@ -736,7 +757,7 @@
         r = Client.disable(self, quiet=quiet)
         if not quiet and oldstate != self.enabled:
             # Emit D-Bus signal
-            self.PropertyChanged(dbus.String(u"enabled"),
+            self.PropertyChanged(dbus.String(u"Enabled"),
                                  dbus.Boolean(False, variant_level=1))
         return r
     
@@ -754,7 +775,7 @@
         self.checker_callback_tag = None
         self.checker = None
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"checker_running"),
+        self.PropertyChanged(dbus.String(u"CheckerRunning"),
                              dbus.Boolean(False, variant_level=1))
         if os.WIFEXITED(condition):
             exitstatus = os.WEXITSTATUS(condition)
@@ -775,7 +796,7 @@
         r = Client.checked_ok(self, *args, **kwargs)
         # Emit D-Bus signal
         self.PropertyChanged(
-            dbus.String(u"last_checked_ok"),
+            dbus.String(u"LastCheckedOK"),
             (self._datetime_to_dbus(self.last_checked_ok,
                                     variant_level=1)))
         return r
@@ -793,7 +814,7 @@
             # Emit D-Bus signal
             self.CheckerStarted(self.current_checker_command)
             self.PropertyChanged(
-                dbus.String(u"checker_running"),
+                dbus.String(u"CheckerRunning"),
                 dbus.Boolean(True, variant_level=1))
         return r
     
@@ -802,7 +823,7 @@
         r = Client.stop_checker(self, *args, **kwargs)
         if (old_checker is not None
             and getattr(self, u"checker", None) is None):
-            self.PropertyChanged(dbus.String(u"checker_running"),
+            self.PropertyChanged(dbus.String(u"CheckerRunning"),
                                  dbus.Boolean(False, variant_level=1))
         return r
 
@@ -811,11 +832,11 @@
         return False
     
     def approve(self, value=True):
+        self.send_changedstate()
         self._approved = value
-        gobject.timeout_add(self._timedelta_to_milliseconds(self.approved_duration, self._reset_approved))
-
-    def approved_pending(self):
-        return self.approvals_pending > 0
+        gobject.timeout_add(self._timedelta_to_milliseconds
+                            (self.approval_duration),
+                            self._reset_approved)
     
     
     ## D-Bus methods, signals & properties
@@ -844,26 +865,23 @@
     # GotSecret - signal
     @dbus.service.signal(_interface)
     def GotSecret(self):
-        "D-Bus signal"
-        if self.approved_pending():
-            self.PropertyChanged(dbus.String(u"checker_running"),
-                                 dbus.Boolean(False, variant_level=1))
+        """D-Bus signal
+        Is sent after a successful transfer of secret from the Mandos
+        server to mandos-client
+        """
+        pass
     
     # Rejected - signal
     @dbus.service.signal(_interface, signature=u"s")
     def Rejected(self, reason):
         "D-Bus signal"
-        if self.approved_pending():
-            self.PropertyChanged(dbus.String(u"checker_running"),
-                                 dbus.Boolean(False, variant_level=1))
+        pass
     
     # NeedApproval - signal
-    @dbus.service.signal(_interface, signature=u"db")
+    @dbus.service.signal(_interface, signature=u"tb")
     def NeedApproval(self, timeout, default):
         "D-Bus signal"
-        if not self.approved_pending():
-            self.PropertyChanged(dbus.String(u"approved_pending"),
-                                 dbus.Boolean(True, variant_level=1))
+        pass
     
     ## Methods
 
@@ -902,67 +920,82 @@
     
     ## Properties
     
-    # approved_pending - property
+    # ApprovalPending - property
     @dbus_service_property(_interface, signature=u"b", access=u"read")
-    def approved_pending_dbus_property(self):
-        return dbus.Boolean(self.approved_pending())
+    def ApprovalPending_dbus_property(self):
+        return dbus.Boolean(bool(self.approvals_pending))
     
-    # approved_by_default - property
+    # ApprovedByDefault - property
     @dbus_service_property(_interface, signature=u"b",
                            access=u"readwrite")
-    def approved_by_default_dbus_property(self):
-        return dbus.Boolean(self.approved_by_default)
-    
-    # approved_delay - property
-    @dbus_service_property(_interface, signature=u"t",
-                           access=u"readwrite")
-    def approved_delay_dbus_property(self):
-        return dbus.UInt64(self.approved_delay_milliseconds())
-    
-    # approved_duration - property
-    @dbus_service_property(_interface, signature=u"t",
-                           access=u"readwrite")
-    def approved_duration_dbus_property(self):
-        return dbus.UInt64(self._timedelta_to_milliseconds(
-                self.approved_duration))
-    
-    # name - property
+    def ApprovedByDefault_dbus_property(self, value=None):
+        if value is None:       # get
+            return dbus.Boolean(self.approved_by_default)
+        self.approved_by_default = bool(value)
+        # Emit D-Bus signal
+        self.PropertyChanged(dbus.String(u"ApprovedByDefault"),
+                             dbus.Boolean(value, variant_level=1))
+    
+    # ApprovalDelay - property
+    @dbus_service_property(_interface, signature=u"t",
+                           access=u"readwrite")
+    def ApprovalDelay_dbus_property(self, value=None):
+        if value is None:       # get
+            return dbus.UInt64(self.approval_delay_milliseconds())
+        self.approval_delay = datetime.timedelta(0, 0, 0, value)
+        # Emit D-Bus signal
+        self.PropertyChanged(dbus.String(u"ApprovalDelay"),
+                             dbus.UInt64(value, variant_level=1))
+    
+    # ApprovalDuration - property
+    @dbus_service_property(_interface, signature=u"t",
+                           access=u"readwrite")
+    def ApprovalDuration_dbus_property(self, value=None):
+        if value is None:       # get
+            return dbus.UInt64(self._timedelta_to_milliseconds(
+                    self.approval_duration))
+        self.approval_duration = datetime.timedelta(0, 0, 0, value)
+        # Emit D-Bus signal
+        self.PropertyChanged(dbus.String(u"ApprovalDuration"),
+                             dbus.UInt64(value, variant_level=1))
+    
+    # Name - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def name_dbus_property(self):
+    def Name_dbus_property(self):
         return dbus.String(self.name)
     
-    # fingerprint - property
+    # Fingerprint - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def fingerprint_dbus_property(self):
+    def Fingerprint_dbus_property(self):
         return dbus.String(self.fingerprint)
     
-    # host - property
+    # Host - property
     @dbus_service_property(_interface, signature=u"s",
                            access=u"readwrite")
-    def host_dbus_property(self, value=None):
+    def Host_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.String(self.host)
         self.host = value
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"host"),
+        self.PropertyChanged(dbus.String(u"Host"),
                              dbus.String(value, variant_level=1))
     
-    # created - property
+    # Created - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def created_dbus_property(self):
+    def Created_dbus_property(self):
         return dbus.String(self._datetime_to_dbus(self.created))
     
-    # last_enabled - property
+    # LastEnabled - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def last_enabled_dbus_property(self):
+    def LastEnabled_dbus_property(self):
         if self.last_enabled is None:
             return dbus.String(u"")
         return dbus.String(self._datetime_to_dbus(self.last_enabled))
     
-    # enabled - property
+    # Enabled - property
     @dbus_service_property(_interface, signature=u"b",
                            access=u"readwrite")
-    def enabled_dbus_property(self, value=None):
+    def Enabled_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.Boolean(self.enabled)
         if value:
@@ -970,10 +1003,10 @@
         else:
             self.disable()
     
-    # last_checked_ok - property
+    # LastCheckedOK - property
     @dbus_service_property(_interface, signature=u"s",
                            access=u"readwrite")
-    def last_checked_ok_dbus_property(self, value=None):
+    def LastCheckedOK_dbus_property(self, value=None):
         if value is not None:
             self.checked_ok()
             return
@@ -982,15 +1015,15 @@
         return dbus.String(self._datetime_to_dbus(self
                                                   .last_checked_ok))
     
-    # timeout - property
+    # Timeout - property
     @dbus_service_property(_interface, signature=u"t",
                            access=u"readwrite")
-    def timeout_dbus_property(self, value=None):
+    def Timeout_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.UInt64(self.timeout_milliseconds())
         self.timeout = datetime.timedelta(0, 0, 0, value)
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"timeout"),
+        self.PropertyChanged(dbus.String(u"Timeout"),
                              dbus.UInt64(value, variant_level=1))
         if getattr(self, u"disable_initiator_tag", None) is None:
             return
@@ -1010,15 +1043,15 @@
             self.disable_initiator_tag = (gobject.timeout_add
                                           (time_to_die, self.disable))
     
-    # interval - property
+    # Interval - property
     @dbus_service_property(_interface, signature=u"t",
                            access=u"readwrite")
-    def interval_dbus_property(self, value=None):
+    def Interval_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.UInt64(self.interval_milliseconds())
         self.interval = datetime.timedelta(0, 0, 0, value)
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"interval"),
+        self.PropertyChanged(dbus.String(u"Interval"),
                              dbus.UInt64(value, variant_level=1))
         if getattr(self, u"checker_initiator_tag", None) is None:
             return
@@ -1028,22 +1061,22 @@
                                       (value, self.start_checker))
         self.start_checker()    # Start one now, too
 
-    # checker - property
+    # Checker - property
     @dbus_service_property(_interface, signature=u"s",
                            access=u"readwrite")
-    def checker_dbus_property(self, value=None):
+    def Checker_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.String(self.checker_command)
         self.checker_command = value
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"checker"),
+        self.PropertyChanged(dbus.String(u"Checker"),
                              dbus.String(self.checker_command,
                                          variant_level=1))
     
-    # checker_running - property
+    # CheckerRunning - property
     @dbus_service_property(_interface, signature=u"b",
                            access=u"readwrite")
-    def checker_running_dbus_property(self, value=None):
+    def CheckerRunning_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.Boolean(self.checker is not None)
         if value:
@@ -1051,15 +1084,15 @@
         else:
             self.stop_checker()
     
-    # object_path - property
+    # ObjectPath - property
     @dbus_service_property(_interface, signature=u"o", access=u"read")
-    def object_path_dbus_property(self):
+    def ObjectPath_dbus_property(self):
         return self.dbus_object_path # is already a dbus.ObjectPath
     
-    # secret = property
+    # Secret = property
     @dbus_service_property(_interface, signature=u"ay",
                            access=u"write", byte_arrays=True)
-    def secret_dbus_property(self, value):
+    def Secret_dbus_property(self, value):
         self.secret = str(value)
     
     del _interface
@@ -1163,8 +1196,8 @@
                 except KeyError:
                     return
                 
-                if client.approved_delay:
-                    delay = client.approved_delay
+                if client.approval_delay:
+                    delay = client.approval_delay
                     client.approvals_pending += 1
                     approval_required = True
                 
@@ -1177,23 +1210,23 @@
                             client.Rejected("Disabled")                    
                         return
                     
-                    if client._approved or not client.approved_delay:
+                    if client._approved or not client.approval_delay:
                         #We are approved or approval is disabled
                         break
                     elif client._approved is None:
-                        logger.info(u"Client %s need approval",
+                        logger.info(u"Client %s needs approval",
                                     client.name)
                         if self.server.use_dbus:
                             # Emit D-Bus signal
                             client.NeedApproval(
-                                client.approved_delay_milliseconds(),
+                                client.approval_delay_milliseconds(),
                                 client.approved_by_default)
                     else:
                         logger.warning(u"Client %s was not approved",
                                        client.name)
                         if self.server.use_dbus:
                             # Emit D-Bus signal
-                            client.Rejected("Disapproved")
+                            client.Rejected("Denied")
                         return
                     
                     #wait until timeout or approved
@@ -1210,7 +1243,7 @@
                                            client.name)
                             if self.server.use_dbus:
                                 # Emit D-Bus signal
-                                client.Rejected("Time out")
+                                client.Rejected("Approval timed out")
                             return
                         else:
                             break
@@ -1219,8 +1252,11 @@
                 
                 sent_size = 0
                 while sent_size < len(client.secret):
-                    # XXX handle session exception
-                    sent = session.send(client.secret[sent_size:])
+                    try:
+                        sent = session.send(client.secret[sent_size:])
+                    except (gnutls.errors.GNUTLSError), error:
+                        logger.warning("gnutls send failed")
+                        return
                     logger.debug(u"Sent: %d, remaining: %d",
                                  sent, len(client.secret)
                                  - (sent_size + sent))
@@ -1236,7 +1272,10 @@
             finally:
                 if approval_required:
                     client.approvals_pending -= 1
-                session.bye()
+                try:
+                    session.bye()
+                except (gnutls.errors.GNUTLSError), error:
+                    logger.warning("GnuTLS bye failed")
     
     @staticmethod
     def peer_certificate(session):
@@ -1327,7 +1366,9 @@
 
         super(MultiprocessingMixInWithPipe,
               self).process_request(request, client_address)
+        self.child_pipe.close()
         self.add_pipe(parent_pipe)
+
     def add_pipe(self, parent_pipe):
         """Dummy function; override as necessary"""
         pass
@@ -1446,8 +1487,9 @@
                                        for cond, name in
                                        condition_names.iteritems()
                                        if cond & condition)
-        logger.debug(u"Handling IPC: FD = %d, condition = %s", source,
-                     conditions_string)
+        # error or the other end of multiprocessing.Pipe has closed
+        if condition & (gobject.IO_ERR | condition & gobject.IO_HUP):
+            return False
         
         # Read a request from the child
         request = parent_pipe.recv()
@@ -1491,12 +1533,12 @@
                 parent_pipe.send(('function',))
             else:
                 parent_pipe.send(('data', client_object.__getattribute__(attrname)))
-
+        
         if command == 'setattr':
             attrname = request[1]
             value = request[2]
             setattr(client_object, attrname, value)
-            
+
         return True
 
 
@@ -1605,6 +1647,8 @@
     parser.add_option("--debug", action=u"store_true",
                       help=u"Debug mode; run in foreground and log to"
                       u" terminal")
+    parser.add_option("--debuglevel", type=u"string", metavar="Level",
+                      help=u"Debug level for stdout output")
     parser.add_option("--priority", type=u"string", help=u"GnuTLS"
                       u" priority string (see GnuTLS documentation)")
     parser.add_option("--servicename", type=u"string",
@@ -1635,6 +1679,7 @@
                         u"servicename": u"Mandos",
                         u"use_dbus": u"True",
                         u"use_ipv6": u"True",
+                        u"debuglevel": u"",
                         }
     
     # Parse config file for server-global settings
@@ -1657,7 +1702,7 @@
     # options, if set.
     for option in (u"interface", u"address", u"port", u"debug",
                    u"priority", u"servicename", u"configdir",
-                   u"use_dbus", u"use_ipv6"):
+                   u"use_dbus", u"use_ipv6", u"debuglevel"):
         value = getattr(options, option)
         if value is not None:
             server_settings[option] = value
@@ -1672,13 +1717,10 @@
     
     # For convenience
     debug = server_settings[u"debug"]
+    debuglevel = server_settings[u"debuglevel"]
     use_dbus = server_settings[u"use_dbus"]
     use_ipv6 = server_settings[u"use_ipv6"]
-    
-    if not debug:
-        syslogger.setLevel(logging.WARNING)
-        console.setLevel(logging.WARNING)
-    
+
     if server_settings[u"servicename"] != u"Mandos":
         syslogger.setFormatter(logging.Formatter
                                (u'Mandos (%s) [%%(process)d]:'
@@ -1690,8 +1732,8 @@
                         u"interval": u"5m",
                         u"checker": u"fping -q -- %%(host)s",
                         u"host": u"",
-                        u"approved_delay": u"0s",
-                        u"approved_duration": u"1s",
+                        u"approval_delay": u"0s",
+                        u"approval_duration": u"1s",
                         }
     client_config = configparser.SafeConfigParser(client_defaults)
     client_config.read(os.path.join(server_settings[u"configdir"],
@@ -1703,7 +1745,8 @@
     tcp_server = MandosServer((server_settings[u"address"],
                                server_settings[u"port"]),
                               ClientHandler,
-                              interface=server_settings[u"interface"],
+                              interface=(server_settings[u"interface"]
+                                         or None),
                               use_ipv6=use_ipv6,
                               gnutls_priority=
                               server_settings[u"priority"],
@@ -1735,8 +1778,17 @@
         if error[0] != errno.EPERM:
             raise error
     
-    # Enable all possible GnuTLS debugging
+    if not debug and not debuglevel:
+        syslogger.setLevel(logging.WARNING)
+        console.setLevel(logging.WARNING)
+    if debuglevel:
+        level = getattr(logging, debuglevel.upper())
+        syslogger.setLevel(level)
+        console.setLevel(level)
+
     if debug:
+        # Enable all possible GnuTLS debugging
+        
         # "Use a log level over 10 to enable all debugging options."
         # - GnuTLS manual
         gnutls.library.functions.gnutls_global_set_log_level(11)
@@ -1747,6 +1799,16 @@
         
         (gnutls.library.functions
          .gnutls_global_set_log_function(debug_gnutls))
+        
+        # Redirect stdin so all checkers get /dev/null
+        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
+        os.dup2(null, sys.stdin.fileno())
+        if null > 2:
+            os.close(null)
+    else:
+        # No console logging
+        logger.removeHandler(console)
+    
     
     global main_loop
     # From the Avahi example code
@@ -1770,6 +1832,13 @@
     if server_settings["interface"]:
         service.interface = (if_nametoindex
                              (str(server_settings[u"interface"])))
+
+    if not debug:
+        # Close all input and output, do double fork, etc.
+        daemon()
+        
+    global multiprocessing_manager
+    multiprocessing_manager = multiprocessing.Manager()
     
     client_class = Client
     if use_dbus:
@@ -1793,19 +1862,7 @@
             for section in client_config.sections()))
     if not tcp_server.clients:
         logger.warning(u"No clients defined")
-    
-    if debug:
-        # Redirect stdin so all checkers get /dev/null
-        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
-        os.dup2(null, sys.stdin.fileno())
-        if null > 2:
-            os.close(null)
-    else:
-        # No console logging
-        logger.removeHandler(console)
-        # Close all input and output, do double fork, etc.
-        daemon()
-    
+        
     try:
         with pidfile:
             pid = os.getpid()

=== modified file 'mandos-clients.conf.xml'
--- mandos-clients.conf.xml	2009-12-25 23:13:47 +0000
+++ mandos-clients.conf.xml	2010-09-12 03:00:40 +0000
@@ -3,7 +3,7 @@
 	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY CONFNAME "mandos-clients.conf">
 <!ENTITY CONFPATH "<filename>/etc/mandos/clients.conf</filename>">
-<!ENTITY TIMESTAMP "2009-12-09">
+<!ENTITY TIMESTAMP "2010-09-12">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -266,6 +266,58 @@
 	</listitem>
       </varlistentry>
       
+      <varlistentry>
+	<term><option>approved_by_default<literal> = </literal
+          >{ <literal >1</literal> | <literal>yes</literal> | <literal
+	  >true</literal> | <literal>on</literal> | <literal
+	  >0</literal> | <literal>no</literal> | <literal
+	  >false</literal> | <literal>off</literal> }</option></term>
+	<listitem>
+	  <para>
+	    Whether to approve a client by default after
+	    the <option>approval_delay</option>.  The default
+	    is <quote>True</quote>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>approval_delay<literal> = </literal><replaceable
+	>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is <emphasis>optional</emphasis>.
+	  </para>
+	  <para>
+	    How long to wait for external approval before resorting to
+	    use the <option>approved_by_default</option> value.  The
+	    default is <quote>0s</quote>, i.e. not to wait.
+	  </para>
+	  <para>
+	    The format of <replaceable>TIME</replaceable> is the same
+	    as for <varname>timeout</varname> above.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>approval_duration<literal> = </literal
+        ><replaceable>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is <emphasis>optional</emphasis>.
+	  </para>
+	  <para>
+	    How long an external approval lasts.  The default is 1
+	    second.
+	  </para>
+	  <para>
+	    The format of <replaceable>TIME</replaceable> is the same
+	    as for <varname>timeout</varname> above.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
     </variablelist>
   </refsect1>
   
@@ -379,6 +431,8 @@
 fingerprint = 3e393aeaefb84c7e89e2f547b3a107558fca3a27
 secfile = /etc/mandos/bar-secret
 timeout = 15m
+approved_by_default = False
+approval_delay = 30s
       </programlisting>
     </informalexample>
   </refsect1>

=== modified file 'mandos-ctl'
--- mandos-ctl	2009-12-26 01:19:47 +0000
+++ mandos-ctl	2010-09-12 18:23:40 +0000
@@ -8,34 +8,30 @@
 import locale
 import datetime
 import re
+import os
 
 locale.setlocale(locale.LC_ALL, u'')
 
 tablewords = {
-    'name': u'Name',
-    'enabled': u'Enabled',
-    'timeout': u'Timeout',
-    'last_checked_ok': u'Last Successful Check',
-    'created': u'Created',
-    'interval': u'Interval',
-    'host': u'Host',
-    'fingerprint': u'Fingerprint',
-    'checker_running': u'Check Is Running',
-    'last_enabled': u'Last Enabled',
-    'checker': u'Checker',
+    'Name': u'Name',
+    'Enabled': u'Enabled',
+    'Timeout': u'Timeout',
+    'LastCheckedOK': u'Last Successful Check',
+    'Created': u'Created',
+    'Interval': u'Interval',
+    'Host': u'Host',
+    'Fingerprint': u'Fingerprint',
+    'CheckerRunning': u'Check Is Running',
+    'LastEnabled': u'Last Enabled',
+    'Checker': u'Checker',
     }
-defaultkeywords = ('name', 'enabled', 'timeout', 'last_checked_ok')
+defaultkeywords = ('Name', 'Enabled', 'Timeout', 'LastCheckedOK')
 domain = 'se.bsnet.fukt'
 busname = domain + '.Mandos'
 server_path = '/'
 server_interface = domain + '.Mandos'
 client_interface = domain + '.Mandos.Client'
 version = "1.0.14"
-bus = dbus.SystemBus()
-mandos_dbus_objc = bus.get_object(busname, server_path)
-mandos_serv = dbus.Interface(mandos_dbus_objc,
-                             dbus_interface = server_interface)
-mandos_clients = mandos_serv.GetAllClientsWithProperties()
 
 def timedelta_to_milliseconds(td):
     "Convert a datetime.timedelta object to milliseconds"
@@ -93,7 +89,7 @@
         timevalue += delta
     return timevalue
 
-def print_clients(clients):
+def print_clients(clients, keywords):
     def valuetostring(value, keyword):
         if type(value) is dbus.Boolean:
             return u"Yes" if value else u"No"
@@ -104,7 +100,8 @@
     # Create format string to print table rows
     format_string = u' '.join(u'%%-%ds' %
                               max(len(tablewords[key]),
-                                  max(len(valuetostring(client[key], key))
+                                  max(len(valuetostring(client[key],
+                                                        key))
                                       for client in
                                       clients))
                               for key in keywords)
@@ -113,92 +110,160 @@
     for client in clients:
         print format_string % tuple(valuetostring(client[key], key)
                                     for key in keywords)
-
-parser = OptionParser(version = "%%prog %s" % version)
-parser.add_option("-a", "--all", action="store_true",
-                  help="Print all fields")
-parser.add_option("-e", "--enable", action="store_true",
-                  help="Enable client")
-parser.add_option("-d", "--disable", action="store_true",
-                  help="disable client")
-parser.add_option("-b", "--bump-timeout", action="store_true",
-                  help="Bump timeout for client")
-parser.add_option("--start-checker", action="store_true",
-                  help="Start checker for client")
-parser.add_option("--stop-checker", action="store_true",
-                  help="Stop checker for client")
-parser.add_option("-V", "--is-enabled", action="store_true",
-                  help="Check if client is enabled")
-parser.add_option("-r", "--remove", action="store_true",
-                  help="Remove client")
-parser.add_option("-c", "--checker", type="string",
-                  help="Set checker command for client")
-parser.add_option("-t", "--timeout", type="string",
-                  help="Set timeout for client")
-parser.add_option("-i", "--interval", type="string",
-                  help="Set checker interval for client")
-parser.add_option("-H", "--host", type="string",
-                  help="Set host for client")
-parser.add_option("-s", "--secret", type="string",
-                  help="Set password blob (file) for client")
-options, client_names = parser.parse_args()
-
-# Compile list of clients to process
-clients=[]
-for name in client_names:
-    for path, client in mandos_clients.iteritems():
-        if client['name'] == name:
-            client_objc = bus.get_object(busname, path)
-            clients.append(client_objc)
-            break
-    else:
-        print >> sys.stderr, "Client not found on server: %r" % name
-        sys.exit(1)
-
-if not clients and mandos_clients.values():
-    keywords = defaultkeywords
-    if options.all:
-        keywords = ('name', 'enabled', 'timeout', 'last_checked_ok',
-                    'created', 'interval', 'host', 'fingerprint',
-                    'checker_running', 'last_enabled', 'checker')
-    print_clients(mandos_clients.values())
-
-# Process each client in the list by all selected options
-for client in clients:
-    if options.remove:
-        mandos_serv.RemoveClient(client.__dbus_object_path__)
-    if options.enable:
-        client.Enable(dbus_interface=client_interface)
-    if options.disable:
-        client.Disable(dbus_interface=client_interface)
-    if options.bump_timeout:
-        client.CheckedOK(dbus_interface=client_interface)
-    if options.start_checker:
-        client.StartChecker(dbus_interface=client_interface)
-    if options.stop_checker:
-        client.StopChecker(dbus_interface=client_interface)
-    if options.is_enabled:
-        sys.exit(0 if client.Get(client_interface,
-                                 u"enabled",
-                                 dbus_interface=dbus.PROPERTIES_IFACE)
-                 else 1)
-    if options.checker:
-        client.Set(client_interface, u"checker", options.checker,
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.host:
-        client.Set(client_interface, u"host", options.host,
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.interval:
-        client.Set(client_interface, u"interval",
-                   timedelta_to_milliseconds
-                   (string_to_delta(options.interval)),
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.timeout:
-        client.Set(client_interface, u"timeout",
-                   timedelta_to_milliseconds(string_to_delta
-                                             (options.timeout)),
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.secret:
-        client.Set(client_interface, u"secret",
-                   dbus.ByteArray(open(options.secret, u'rb').read()),
-                   dbus_interface=dbus.PROPERTIES_IFACE)
+def has_actions(options):
+    return any((options.enable,
+                options.disable,
+                options.bump_timeout,
+                options.start_checker,
+                options.stop_checker,
+                options.is_enabled,
+                options.remove,
+                options.checker is not None,
+                options.timeout is not None,
+                options.interval is not None,
+                options.host is not None,
+                options.secret is not None,
+                options.approve,
+                options.deny))
+        
+def main():
+	parser = OptionParser(version = "%%prog %s" % version)
+	parser.add_option("-a", "--all", action="store_true",
+	                  help="Select all clients")
+	parser.add_option("-v", "--verbose", action="store_true",
+	                  help="Print all fields")
+	parser.add_option("-e", "--enable", action="store_true",
+	                  help="Enable client")
+	parser.add_option("-d", "--disable", action="store_true",
+	                  help="disable client")
+	parser.add_option("-b", "--bump-timeout", action="store_true",
+	                  help="Bump timeout for client")
+	parser.add_option("--start-checker", action="store_true",
+	                  help="Start checker for client")
+	parser.add_option("--stop-checker", action="store_true",
+	                  help="Stop checker for client")
+	parser.add_option("-V", "--is-enabled", action="store_true",
+	                  help="Check if client is enabled")
+	parser.add_option("-r", "--remove", action="store_true",
+	                  help="Remove client")
+	parser.add_option("-c", "--checker", type="string",
+	                  help="Set checker command for client")
+	parser.add_option("-t", "--timeout", type="string",
+	                  help="Set timeout for client")
+	parser.add_option("-i", "--interval", type="string",
+	                  help="Set checker interval for client")
+	parser.add_option("-H", "--host", type="string",
+	                  help="Set host for client")
+	parser.add_option("-s", "--secret", type="string",
+	                  help="Set password blob (file) for client")
+	parser.add_option("-A", "--approve", action="store_true",
+	                  help="Approve any current client request")
+	parser.add_option("-D", "--deny", action="store_true",
+	                  help="Deny any current client request")
+	options, client_names = parser.parse_args()
+
+        if has_actions(options) and not client_names and not options.all:
+            parser.error('Options requires clients names or --all.')
+        if options.verbose and has_actions(options):
+            parser.error('Verbose option can only be used alone or with --all.')
+        if options.all and not has_actions(options):
+            parser.error('--all requires an action')
+            
+        try:
+            bus = dbus.SystemBus()
+            mandos_dbus_objc = bus.get_object(busname, server_path)
+        except dbus.exceptions.DBusException:
+            print >> sys.stderr, "Could not connect to Mandos server"
+            sys.exit(1)
+    
+        mandos_serv = dbus.Interface(mandos_dbus_objc,
+                                     dbus_interface = server_interface)
+
+        #block stderr since dbus library prints to stderr
+        null = os.open(os.path.devnull, os.O_RDWR)
+        stderrcopy = os.dup(sys.stderr.fileno())
+        os.dup2(null, sys.stderr.fileno())
+        os.close(null)
+        try:
+            try:
+                mandos_clients = mandos_serv.GetAllClientsWithProperties()
+            finally:
+                #restore stderr
+                os.dup2(stderrcopy, sys.stderr.fileno())
+                os.close(stderrcopy)
+        except dbus.exceptions.DBusException, e:
+            print >> sys.stderr, "Access denied: Accessing mandos server through dbus."
+            sys.exit(1)
+            
+	# Compile list of clients to process
+	clients=[]
+
+        if options.all or not client_names:
+            clients = (bus.get_object(busname, path) for path in mandos_clients.iterkeys())
+        else:
+            for name in client_names:
+                for path, client in mandos_clients.iteritems():
+                    if client['Name'] == name:
+                        client_objc = bus.get_object(busname, path)
+                        clients.append(client_objc)
+                        break
+                else:
+                    print >> sys.stderr, "Client not found on server: %r" % name
+                    sys.exit(1)
+            
+	if not has_actions(options) and clients:
+	    if options.verbose:
+	        keywords = ('Name', 'Enabled', 'Timeout', 'LastCheckedOK',
+	                    'Created', 'Interval', 'Host', 'Fingerprint',
+	                    'CheckerRunning', 'LastEnabled', 'Checker')
+            else:
+                keywords = defaultkeywords
+                
+	    print_clients(mandos_clients.values(), keywords)
+        else:
+            # Process each client in the list by all selected options
+            for client in clients:
+                if options.remove:
+                    mandos_serv.RemoveClient(client.__dbus_object_path__)
+                if options.enable:
+                    client.Enable(dbus_interface=client_interface)
+                if options.disable:
+                    client.Disable(dbus_interface=client_interface)
+                if options.bump_timeout:
+                    client.CheckedOK(dbus_interface=client_interface)
+                if options.start_checker:
+                    client.StartChecker(dbus_interface=client_interface)
+                if options.stop_checker:
+                    client.StopChecker(dbus_interface=client_interface)
+                if options.is_enabled:
+                    sys.exit(0 if client.Get(client_interface,
+                                             u"Enabled",
+                                             dbus_interface=dbus.PROPERTIES_IFACE)
+                             else 1)
+                if options.checker:
+                    client.Set(client_interface, u"Checker", options.checker,
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.host:
+                    client.Set(client_interface, u"Host", options.host,
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.interval:
+                    client.Set(client_interface, u"Interval",
+                               timedelta_to_milliseconds
+                               (string_to_delta(options.interval)),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.timeout:
+                    client.Set(client_interface, u"Timeout",
+                               timedelta_to_milliseconds(string_to_delta
+                                                         (options.timeout)),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.secret:
+                    client.Set(client_interface, u"Secret",
+                               dbus.ByteArray(open(options.secret, u'rb').read()),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.approve:
+                    client.Approve(dbus.Boolean(True), dbus_interface=client_interface)
+                if options.deny:
+                    client.Approve(dbus.Boolean(False), dbus_interface=client_interface)
+
+if __name__ == '__main__':
+    main()

=== modified file 'mandos-monitor'
--- mandos-monitor	2010-08-31 18:22:25 +0000
+++ mandos-monitor	2010-09-12 03:08:51 +0000
@@ -40,6 +40,22 @@
 urwid.curses_display.curses.A_UNDERLINE |= (
     urwid.curses_display.curses.A_BLINK)
 
+def isoformat_to_datetime(iso):
+    "Parse an ISO 8601 date string to a datetime.datetime()"
+    if not iso:
+        return None
+    d, t = iso.split(u"T", 1)
+    year, month, day = d.split(u"-", 2)
+    hour, minute, second = t.split(u":", 2)
+    second, fraction = divmod(float(second), 1)
+    return datetime.datetime(int(year),
+                             int(month),
+                             int(day),
+                             int(hour),
+                             int(minute),
+                             int(second),           # Whole seconds
+                             int(fraction*1000000)) # Microseconds
+
 class MandosClientPropertyCache(object):
     """This wraps a Mandos Client D-Bus proxy object, caches the
     properties and calls a hook function when any of them are
@@ -57,8 +73,10 @@
         self.properties.update(
             self.proxy.GetAll(client_interface,
                               dbus_interface = dbus.PROPERTIES_IFACE))
-        super(MandosClientPropertyCache, self).__init__(
-            proxy_object=proxy_object, *args, **kwargs)
+
+        #XXX This break good super behaviour!
+#        super(MandosClientPropertyCache, self).__init__(
+#            *args, **kwargs)
     
     def property_changed(self, property=None, value=None):
         """This is called whenever we get a PropertyChanged signal
@@ -83,6 +101,9 @@
         # Logger
         self.logger = logger
         
+        self._update_timer_callback_tag = None
+        self.last_checker_failed = False
+        
         # The widget shown normally
         self._text_widget = urwid.Text(u"")
         # The widget shown when we have focus
@@ -112,39 +133,67 @@
                                      self.rejected,
                                      client_interface,
                                      byte_arrays=True)
+        last_checked_ok = isoformat_to_datetime(self.properties
+                                                [u"LastCheckedOK"])
+        if last_checked_ok is None:
+            self.last_checker_failed = True
+        else:
+            self.last_checker_failed = ((datetime.datetime.utcnow()
+                                         - last_checked_ok)
+                                        > datetime.timedelta
+                                        (milliseconds=
+                                         self.properties
+                                         [u"Interval"]))
+        if self.last_checker_failed:
+            self._update_timer_callback_tag = (gobject.timeout_add
+                                               (1000,
+                                                self.update_timer))
     
     def checker_completed(self, exitstatus, condition, command):
         if exitstatus == 0:
+            if self.last_checker_failed:
+                self.last_checker_failed = False
+                gobject.source_remove(self._update_timer_callback_tag)
+                self._update_timer_callback_tag = None
             #self.logger(u'Checker for client %s (command "%s")'
             #            u' was successful'
-            #            % (self.properties[u"name"], command))
+            #            % (self.properties[u"Name"], command))
+            self.update()
             return
+        # Checker failed
+        if not self.last_checker_failed:
+            self.last_checker_failed = True
+            self._update_timer_callback_tag = (gobject.timeout_add
+                                               (1000,
+                                                self.update_timer))
         if os.WIFEXITED(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' failed with exit code %s'
-                        % (self.properties[u"name"], command,
+                        % (self.properties[u"Name"], command,
                            os.WEXITSTATUS(condition)))
-            return
-        if os.WIFSIGNALED(condition):
+        elif os.WIFSIGNALED(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' was killed by signal %s'
-                        % (self.properties[u"name"], command,
+                        % (self.properties[u"Name"], command,
                            os.WTERMSIG(condition)))
-            return
-        if os.WCOREDUMP(condition):
+        elif os.WCOREDUMP(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' dumped core'
-                        % (self.properties[u"name"], command))
-        self.logger(u'Checker for client %s completed mysteriously')
+                        % (self.properties[u"Name"], command))
+        else:
+            self.logger(u'Checker for client %s completed'
+                        u' mysteriously')
+        self.update()
     
     def checker_started(self, command):
         #self.logger(u'Client %s started checker "%s"'
-        #            % (self.properties[u"name"], unicode(command)))
+        #            % (self.properties[u"Name"], unicode(command)))
         pass
     
     def got_secret(self):
+        self.last_checker_failed = False
         self.logger(u'Client %s received its secret'
-                    % self.properties[u"name"])
+                    % self.properties[u"Name"])
     
     def need_approval(self, timeout, default):
         if not default:
@@ -152,11 +201,11 @@
         else:
             message = u'Client %s will get its secret in %s seconds'
         self.logger(message
-                    % (self.properties[u"name"], timeout/1000))
+                    % (self.properties[u"Name"], timeout/1000))
     
     def rejected(self, reason):
         self.logger(u'Client %s was rejected; reason: %s'
-                    % (self.properties[u"name"], reason))
+                    % (self.properties[u"Name"], reason))
     
     def selectable(self):
         """Make this a "selectable" widget.
@@ -184,14 +233,35 @@
                           u"bold-underline-blink":
                               u"bold-underline-blink-standout",
                           }
-        
+
         # Rebuild focus and non-focus widgets using current properties
-        self._text = (u'%(name)s: %(enabled)s'
-                      % { u"name": self.properties[u"name"],
-                          u"enabled":
-                              (u"enabled"
-                               if self.properties[u"enabled"]
-                               else u"DISABLED")})
+
+        # Base part of a client. Name!
+        base = (u'%(name)s: '
+                      % {u"name": self.properties[u"Name"]})
+        if not self.properties[u"Enabled"]:
+            message = u"DISABLED"
+        elif self.properties[u"ApprovalPending"]:
+            if self.properties[u"ApprovedByDefault"]:
+                message = u"Connection established to client. (d)eny?"
+            else:
+                message = u"Seeks approval to send secret. (a)pprove?"
+        elif self.last_checker_failed:
+            timeout = datetime.timedelta(milliseconds
+                                         = self.properties
+                                         [u"Timeout"])
+            last_ok = isoformat_to_datetime(
+                max((self.properties[u"LastCheckedOK"]
+                     or self.properties[u"Created"]),
+                    self.properties[u"LastEnabled"]))
+            timer = timeout - (datetime.datetime.utcnow() - last_ok)
+            message = (u'A checker has failed! Time until client'
+                       u' gets diabled: %s'
+                           % unicode(timer).rsplit(".", 1)[0])
+        else:
+            message = u"enabled"
+        self._text = "%s%s" % (base, message)
+            
         if not urwid.supports_unicode():
             self._text = self._text.encode("ascii", "replace")
         textlist = [(u"normal", self._text)]
@@ -208,7 +278,15 @@
         if self.update_hook is not None:
             self.update_hook()
     
+    def update_timer(self):
+        "called by gobject"
+        self.update()
+        return True             # Keep calling this
+    
     def delete(self):
+        if self._update_timer_callback_tag is not None:
+            gobject.source_remove(self._update_timer_callback_tag)
+            self._update_timer_callback_tag = None
         if self.delete_hook is not None:
             self.delete_hook(self)
     
@@ -221,19 +299,25 @@
     def keypress(self, (maxcol,), key):
         """Handle keys.
         This overrides the method from urwid.FlowWidget"""
-        if key == u"e" or key == u"+":
-            self.proxy.Enable()
-        elif key == u"d" or key == u"-":
-            self.proxy.Disable()
+        if key == u"+":
+            self.proxy.Enable(dbus_interface = client_interface)
+        elif key == u"-":
+            self.proxy.Disable(dbus_interface = client_interface)
+        elif key == u"a":
+            self.proxy.Approve(dbus.Boolean(True, variant_level=1),
+                               dbus_interface = client_interface)
+        elif key == u"d":
+            self.proxy.Approve(dbus.Boolean(False, variant_level=1),
+                                  dbus_interface = client_interface)
         elif key == u"r" or key == u"_" or key == u"ctrl k":
             self.server_proxy_object.RemoveClient(self.proxy
                                                   .object_path)
         elif key == u"s":
-            self.proxy.StartChecker()
+            self.proxy.StartChecker(dbus_interface = client_interface)
         elif key == u"S":
-            self.proxy.StopChecker()
+            self.proxy.StopChecker(dbus_interface = client_interface)
         elif key == u"C":
-            self.proxy.CheckedOK()
+            self.proxy.CheckedOK(dbus_interface = client_interface)
         # xxx
 #         elif key == u"p" or key == "=":
 #             self.proxy.pause()
@@ -241,10 +325,10 @@
 #             self.proxy.unpause()
 #         elif key == u"RET":
 #             self.open()
-        elif key == u"+":
-            self.proxy.Approve(True)
-        elif key == u"-":
-            self.proxy.Approve(False)
+#        elif key == u"+":
+#            self.proxy.Approve(True)
+#        elif key == u"-":
+#            self.proxy.Approve(False)
         else:
             return key
     
@@ -266,7 +350,8 @@
     use them as an excuse to shift focus away from this widget.
     """
     def keypress(self, (maxcol, maxrow), key):
-        ret = super(ConstrainedListBox, self).keypress((maxcol, maxrow), key)
+        ret = super(ConstrainedListBox, self).keypress((maxcol,
+                                                        maxrow), key)
         if ret in (u"up", u"down"):
             return
         return ret
@@ -389,10 +474,13 @@
         Call this when the widget layout needs to change"""
         self.uilist = []
         #self.uilist.append(urwid.ListBox(self.clients))
-        self.uilist.append(urwid.Frame(ConstrainedListBox(self.clients),
+        self.uilist.append(urwid.Frame(ConstrainedListBox(self.
+                                                          clients),
                                        #header=urwid.Divider(),
                                        header=None,
-                                       footer=urwid.Divider(div_char=self.divider)))
+                                       footer=
+                                       urwid.Divider(div_char=
+                                                     self.divider)))
         if self.log_visible:
             self.uilist.append(self.logbox)
             pass
@@ -461,7 +549,7 @@
         if path is None:
             path = client.proxy.object_path
         self.clients_dict[path] = client
-        self.clients.sort(None, lambda c: c.properties[u"name"])
+        self.clients.sort(None, lambda c: c.properties[u"Name"])
         self.refresh()
     
     def remove_client(self, client, path=None):
@@ -542,12 +630,14 @@
                 self.log_message_raw((u"bold",
                                       u"  "
                                       .join((u"Clients:",
-                                             u"e: Enable",
-                                             u"d: Disable",
+                                             u"+: Enable",
+                                             u"-: Disable",
                                              u"r: Remove",
                                              u"s: Start new checker",
                                              u"S: Stop checker",
-                                             u"C: Checker OK"))))
+                                             u"C: Checker OK",
+                                             u"a: Approve",
+                                             u"d: Deny"))))
                 self.refresh()
             elif key == u"tab":
                 if self.topwidget.get_focus() is self.logbox:
@@ -581,6 +671,8 @@
 ui = UserInterface()
 try:
     ui.run()
+except KeyboardInterrupt:
+    ui.screen.stop()
 except Exception, e:
     ui.log_message(unicode(e))
     ui.screen.stop()

=== modified file 'mandos.lsm'
--- mandos.lsm	2009-10-26 21:16:16 +0000
+++ mandos.lsm	2010-09-12 03:22:48 +0000
@@ -1,7 +1,7 @@
 Begin4
 Title:		Mandos
 Version:	1.0.14
-Entered-date:	2009-10-25
+Entered-date:	2010-09-12
 Description:	The Mandos system allows computers to have encrypted
 root file systems and at the same time be capable of remote and/or
 unattended reboots.

=== modified file 'mandos.xml'
--- mandos.xml	2009-12-25 23:13:47 +0000
+++ mandos.xml	2010-09-12 03:00:40 +0000
@@ -2,7 +2,7 @@
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 "http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY COMMANDNAME "mandos">
-<!ENTITY TIMESTAMP "2009-12-09">
+<!ENTITY TIMESTAMP "2010-09-11">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -353,8 +353,8 @@
     <para>
       The server will by default provide a D-Bus system bus interface.
       This interface will only be accessible by the root user or a
-      Mandos-specific user, if such a user exists.
-      <!-- XXX -->
+      Mandos-specific user, if such a user exists.  For documentation
+      of the D-Bus API, see the file <filename>DBUS-API</filename>.
     </para>
   </refsect1>
   
@@ -458,11 +458,6 @@
       has some security implications, see <xref linkend="clients"/>.
     </para>
     <para>
-      There is currently no way of querying the server of the current
-      status of clients, other than analyzing its <systemitem
-      class="service">syslog</systemitem> output.
-    </para>
-    <para>
       There is no fine-grained control over logging and debug output.
     </para>
     <para>

=== modified file 'plugin-runner.c'
--- plugin-runner.c	2010-03-27 18:39:02 +0000
+++ plugin-runner.c	2010-09-09 22:06:10 +0000
@@ -28,7 +28,7 @@
 #include <stdlib.h>		/* malloc(), exit(), EXIT_SUCCESS,
 				   realloc() */
 #include <stdbool.h>		/* bool, true, false */
-#include <stdio.h>		/* perror, fileno(), fprintf(),
+#include <stdio.h>		/* fileno(), fprintf(),
 				   stderr, STDOUT_FILENO */
 #include <sys/types.h>	        /* DIR, fdopendir(), stat(), struct
 				   stat, waitpid(), WIFEXITED(),
@@ -70,6 +70,8 @@
 #include <inttypes.h>		/* intmax_t, PRIdMAX, strtoimax() */
 #include <sysexits.h>		/* EX_OSERR, EX_USAGE, EX_IOERR,
 				   EX_CONFIG, EX_UNAVAILABLE, EX_OK */
+#include <errno.h> 		/* errno */
+#include <error.h>		/* error() */
 
 #define BUFFER_SIZE 256
 
@@ -266,7 +268,7 @@
 	/* No child processes */
 	break;
       }
-      perror("waitpid");
+      error(0, errno, "waitpid");
     }
     
     /* A child exited, find it in process_list */
@@ -357,13 +359,13 @@
   sigemptyset(&sigchld_action.sa_mask);
   ret = sigaddset(&sigchld_action.sa_mask, SIGCHLD);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     exitstatus = EX_OSERR;
     goto fallback;
   }
   ret = sigaction(SIGCHLD, &sigchld_action, &old_sigchld_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     exitstatus = EX_OSERR;
     goto fallback;
   }
@@ -599,7 +601,7 @@
   case ENOMEM:
   default:
     errno = ret;
-    perror("argp_parse");
+    error(0, errno, "argp_parse");
     exitstatus = EX_OSERR;
     goto fallback;
   case EINVAL:
@@ -626,7 +628,7 @@
     custom_argc = 1;
     custom_argv = malloc(sizeof(char*) * 2);
     if(custom_argv == NULL){
-      perror("malloc");
+      error(0, errno, "malloc");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -649,7 +651,7 @@
 	}
 	new_arg = strdup(p);
 	if(new_arg == NULL){
-	  perror("strdup");
+	  error(0, errno, "strdup");
 	  exitstatus = EX_OSERR;
 	  free(org_line);
 	  goto fallback;
@@ -659,7 +661,7 @@
 	custom_argv = realloc(custom_argv, sizeof(char *)
 			      * ((unsigned int) custom_argc + 1));
 	if(custom_argv == NULL){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  exitstatus = EX_OSERR;
 	  free(org_line);
 	  goto fallback;
@@ -672,7 +674,7 @@
       ret = fclose(conffp);
     } while(ret == EOF and errno == EINTR);
     if(ret == EOF){
-      perror("fclose");
+      error(0, errno, "fclose");
       exitstatus = EX_IOERR;
       goto fallback;
     }
@@ -681,7 +683,7 @@
     /* Check for harmful errors and go to fallback. Other errors might
        not affect opening plugins */
     if(errno == EMFILE or errno == ENFILE or errno == ENOMEM){
-      perror("fopen");
+      error(0, errno, "fopen");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -698,7 +700,7 @@
     case ENOMEM:
     default:
       errno = ret;
-      perror("argp_parse");
+      error(0, errno, "argp_parse");
       exitstatus = EX_OSERR;
       goto fallback;
     case EINVAL:
@@ -718,7 +720,7 @@
   case ENOMEM:
   default:
     errno = ret;
-    perror("argp_parse");
+    error(0, errno, "argp_parse");
     exitstatus = EX_OSERR;
     goto fallback;
   case EINVAL:
@@ -743,11 +745,11 @@
   /* Strip permissions down to nobody */
   setgid(gid);
   if(ret == -1){
-    perror("setgid");
+    error(0, errno, "setgid");
   }
   ret = setuid(uid);
   if(ret == -1){
-    perror("setuid");
+    error(0, errno, "setuid");
   }
   
   /* Open plugin directory with close_on_exec flag */
@@ -771,7 +773,7 @@
 		    );
     }
     if(dir_fd == -1){
-      perror("Could not open plugin dir");
+      error(0, errno, "Could not open plugin dir");
       exitstatus = EX_UNAVAILABLE;
       goto fallback;
     }
@@ -780,7 +782,7 @@
   /* Set the FD_CLOEXEC flag on the directory */
     ret = set_cloexec_flag(dir_fd);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
@@ -789,7 +791,7 @@
     
     dir = fdopendir(dir_fd);
     if(dir == NULL){
-      perror("Could not open plugin dir");
+      error(0, errno, "Could not open plugin dir");
       TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
@@ -807,7 +809,7 @@
     /* All directory entries have been processed */
     if(dirst == NULL){
       if(errno == EBADF){
-	perror("readdir");
+	error(0, errno, "readdir");
 	exitstatus = EX_IOERR;
 	goto fallback;
       }
@@ -870,13 +872,13 @@
 					     dirst->d_name));
     }
     if(ret < 0){
-      perror("asprintf");
+      error(0, errno, "asprintf");
       continue;
     }
     
     ret = (int)TEMP_FAILURE_RETRY(stat(filename, &st));
     if(ret == -1){
-      perror("stat");
+      error(0, errno, "stat");
       free(filename);
       continue;
     }
@@ -894,7 +896,7 @@
     
     plugin *p = getplugin(dirst->d_name);
     if(p == NULL){
-      perror("getplugin");
+      error(0, errno, "getplugin");
       free(filename);
       continue;
     }
@@ -912,13 +914,13 @@
       if(g != NULL){
 	for(char **a = g->argv + 1; *a != NULL; a++){
 	  if(not add_argument(p, *a)){
-	    perror("add_argument");
+	    error(0, errno, "add_argument");
 	  }
 	}
 	/* Add global environment variables */
 	for(char **e = g->environ; *e != NULL; e++){
 	  if(not add_environment(p, *e, false)){
-	    perror("add_environment");
+	    error(0, errno, "add_environment");
 	  }
 	}
       }
@@ -929,7 +931,7 @@
     if(p->environ[0] != NULL){
       for(char **e = environ; *e != NULL; e++){
 	if(not add_environment(p, *e, false)){
-	  perror("add_environment");
+	  error(0, errno, "add_environment");
 	}
       }
     }
@@ -937,20 +939,20 @@
     int pipefd[2];
     ret = (int)TEMP_FAILURE_RETRY(pipe(pipefd));
     if(ret == -1){
-      perror("pipe");
+      error(0, errno, "pipe");
       exitstatus = EX_OSERR;
       goto fallback;
     }
     /* Ask OS to automatic close the pipe on exec */
     ret = set_cloexec_flag(pipefd[0]);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       exitstatus = EX_OSERR;
       goto fallback;
     }
     ret = set_cloexec_flag(pipefd[1]);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -959,7 +961,7 @@
 					      &sigchld_action.sa_mask,
 					      NULL));
     if(ret < 0){
-      perror("sigprocmask");
+      error(0, errno, "sigprocmask");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -969,7 +971,7 @@
       pid = fork();
     } while(pid == -1 and errno == EINTR);
     if(pid == -1){
-      perror("fork");
+      error(0, errno, "fork");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -977,18 +979,18 @@
       /* this is the child process */
       ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
       if(ret < 0){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	_exit(EX_OSERR);
       }
       ret = sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
       if(ret < 0){
-	perror("sigprocmask");
+	error(0, errno, "sigprocmask");
 	_exit(EX_OSERR);
       }
       
       ret = dup2(pipefd[1], STDOUT_FILENO); /* replace our stdout */
       if(ret == -1){
-	perror("dup2");
+	error(0, errno, "dup2");
 	_exit(EX_OSERR);
       }
       
@@ -999,12 +1001,12 @@
       }
       if(p->environ[0] == NULL){
 	if(execv(filename, p->argv) < 0){
-	  perror("execv");
+	  error(0, errno, "execv for %s", filename);
 	  _exit(EX_OSERR);
 	}
       } else {
 	if(execve(filename, p->argv, p->environ) < 0){
-	  perror("execve");
+	  error(0, errno, "execve for %s", filename);
 	  _exit(EX_OSERR);
 	}
       }
@@ -1016,12 +1018,12 @@
     free(filename);
     plugin *new_plugin = getplugin(dirst->d_name);
     if(new_plugin == NULL){
-      perror("getplugin");
+      error(0, errno, "getplugin");
       ret = (int)(TEMP_FAILURE_RETRY
 		  (sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask,
 			       NULL)));
       if(ret < 0){
-        perror("sigprocmask");
+        error(0, errno, "sigprocmask");
       }
       exitstatus = EX_OSERR;
       goto fallback;
@@ -1036,7 +1038,7 @@
 					      &sigchld_action.sa_mask,
 					      NULL));
     if(ret < 0){
-      perror("sigprocmask");
+      error(0, errno, "sigprocmask");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -1069,7 +1071,7 @@
     fd_set rfds = rfds_all;
     int select_ret = select(maxfd+1, &rfds, NULL, NULL, NULL);
     if(select_ret == -1 and errno != EINTR){
-      perror("select");
+      error(0, errno, "select");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -1111,7 +1113,7 @@
 					 &sigchld_action.sa_mask,
 					 NULL));
 	  if(ret < 0){
-	    perror("sigprocmask");
+	    error(0, errno, "sigprocmask");
 	    exitstatus = EX_OSERR;
 	    goto fallback;
 	  }
@@ -1125,7 +1127,7 @@
 		      (sigprocmask(SIG_UNBLOCK,
 				   &sigchld_action.sa_mask, NULL)));
 	  if(ret < 0){
-	    perror("sigprocmask");
+	    error(0, errno, "sigprocmask");
 	    exitstatus = EX_OSERR;
 	    goto fallback;
 	  }
@@ -1142,7 +1144,7 @@
 	bool bret = print_out_password(proc->buffer,
 				       proc->buffer_length);
 	if(not bret){
-	  perror("print_out_password");
+	  error(0, errno, "print_out_password");
 	  exitstatus = EX_IOERR;
 	}
 	goto fallback;
@@ -1161,7 +1163,7 @@
 	proc->buffer = realloc(proc->buffer, proc->buffer_size
 			       + (size_t) BUFFER_SIZE);
 	if(proc->buffer == NULL){
-	  perror("malloc");
+	  error(0, errno, "malloc");
 	  exitstatus = EX_OSERR;
 	  goto fallback;
 	}
@@ -1204,7 +1206,7 @@
     }
     bret = print_out_password(passwordbuffer, len);
     if(not bret){
-      perror("print_out_password");
+      error(0, errno, "print_out_password");
       exitstatus = EX_IOERR;
     }
   }
@@ -1212,7 +1214,7 @@
   /* Restore old signal handler */
   ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     exitstatus = EX_OSERR;
   }
   
@@ -1234,7 +1236,7 @@
       ret = kill(p->pid, SIGTERM);
       if(ret == -1 and errno != ESRCH){
 	/* Set-uid proccesses might not get closed */
-	perror("kill");
+	error(0, errno, "kill");
       }
     }
   }
@@ -1244,7 +1246,7 @@
     ret = wait(NULL);
   } while(ret >= 0);
   if(errno != ECHILD){
-    perror("wait");
+    error(0, errno, "wait");
   }
   
   free_plugin_list();

=== modified file 'plugins.d/askpass-fifo.c'
--- plugins.d/askpass-fifo.c	2009-10-18 16:09:05 +0000
+++ plugins.d/askpass-fifo.c	2010-09-07 16:48:58 +0000
@@ -31,7 +31,7 @@
 				   ENOENT, EEXIST, EFAULT, EMFILE,
 				   ENFILE, ENOMEM, EBADF, EINVAL, EIO,
 				   EISDIR, EFBIG */
-#include <stdio.h>		/* perror() */
+#include <error.h>		/* error() */
 #include <stdlib.h>		/* EXIT_FAILURE, NULL, size_t, free(),
 				   realloc(), EXIT_SUCCESS */
 #include <fcntl.h>		/* open(), O_RDONLY */
@@ -51,7 +51,7 @@
   ret = mkfifo(passfifo, S_IRUSR | S_IWUSR);
   if(ret == -1){
     int e = errno;
-    perror("mkfifo");
+    error(0, errno, "mkfifo");
     switch(e){
     case EACCES:
     case ENOTDIR:
@@ -73,7 +73,7 @@
   int fifo_fd = open(passfifo, O_RDONLY);
   if(fifo_fd == -1){
     int e = errno;
-    perror("open");
+    error(0, errno, "open");
     switch(e){
     case EACCES:
     case ENOENT:
@@ -101,7 +101,7 @@
       if(buf_len + blocksize > buf_allocated){
 	char *tmp = realloc(buf, buf_allocated + blocksize);
 	if(tmp == NULL){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  free(buf);
 	  return EX_OSERR;
 	}
@@ -113,7 +113,7 @@
 	int e = errno;
 	free(buf);
 	errno = e;
-	perror("read");
+	error(0, errno, "read");
 	switch(e){
 	case EBADF:
 	case EFAULT:
@@ -141,7 +141,7 @@
       int e = errno;
       free(buf);
       errno = e;
-      perror("write");
+      error(0, errno, "write");
       switch(e){
       case EBADF:
       case EFAULT:
@@ -161,7 +161,7 @@
   ret = close(STDOUT_FILENO);
   if(ret == -1){
     int e = errno;
-    perror("close");
+    error(0, errno, "close");
     switch(e){
     case EBADF:
       return EX_OSFILE;

=== modified file 'plugins.d/mandos-client.c'
--- plugins.d/mandos-client.c	2009-11-01 00:10:28 +0000
+++ plugins.d/mandos-client.c	2010-09-12 18:12:11 +0000
@@ -63,7 +63,7 @@
 				   strtoimax() */
 #include <assert.h>		/* assert() */
 #include <errno.h>		/* perror(), errno */
-#include <time.h>		/* nanosleep(), time() */
+#include <time.h>		/* nanosleep(), time(), sleep() */
 #include <net/if.h>		/* ioctl, ifreq, SIOCGIFFLAGS, IFF_UP,
 				   SIOCSIFFLAGS, if_indextoname(),
 				   if_nametoindex(), IF_NAMESIZE */
@@ -127,6 +127,8 @@
 static const char mandos_protocol_version[] = "1";
 const char *argp_program_version = "mandos-client " VERSION;
 const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
+static const char sys_class_net[] = "/sys/class/net";
+char *connect_to = NULL;
 
 /* Used for passing in values through the Avahi callback functions */
 typedef struct {
@@ -688,9 +690,11 @@
     ret = connect(tcp_sd, &to.in, sizeof(to)); /* IPv4 */
   }
   if(ret < 0){
-    int e = errno;
-    perror("connect");
-    errno = e;
+    if ((errno != ECONNREFUSED and errno != ENETUNREACH) or debug){
+      int e = errno;
+      perror("connect");
+      errno = e;
+    }
     goto mandos_end;
   }
   
@@ -1020,6 +1024,101 @@
   errno = old_errno;
 }
 
+/* 
+ * This function determines if a directory entry in /sys/class/net
+ * corresponds to an acceptable network device.
+ * (This function is passed to scandir(3) as a filter function.)
+ */
+int good_interface(const struct dirent *if_entry){
+  ssize_t ssret;
+  char *flagname = NULL;
+  int ret = asprintf(&flagname, "%s/%s/flags", sys_class_net,
+		     if_entry->d_name);
+  if(ret < 0){
+    perror("asprintf");
+    return 0;
+  }
+  if(if_entry->d_name[0] == '.'){
+    return 0;
+  }
+  int flags_fd = (int)TEMP_FAILURE_RETRY(open(flagname, O_RDONLY));
+  if(flags_fd == -1){
+    perror("open");
+    return 0;
+  }
+  typedef short ifreq_flags;	/* ifreq.ifr_flags in netdevice(7) */
+  /* read line from flags_fd */
+  ssize_t to_read = (sizeof(ifreq_flags)*2)+3; /* "0x1003\n" */
+  char *flagstring = malloc((size_t)to_read+1); /* +1 for final \0 */
+  flagstring[(size_t)to_read] = '\0';
+  if(flagstring == NULL){
+    perror("malloc");
+    close(flags_fd);
+    return 0;
+  }
+  while(to_read > 0){
+    ssret = (ssize_t)TEMP_FAILURE_RETRY(read(flags_fd, flagstring,
+					     (size_t)to_read));
+    if(ssret == -1){
+      perror("read");
+      free(flagstring);
+      close(flags_fd);
+      return 0;
+    }
+    to_read -= ssret;
+    if(ssret == 0){
+      break;
+    }
+  }
+  close(flags_fd);
+  intmax_t tmpmax;
+  char *tmp;
+  errno = 0;
+  tmpmax = strtoimax(flagstring, &tmp, 0);
+  if(errno != 0 or tmp == flagstring or (*tmp != '\0'
+					 and not (isspace(*tmp)))
+     or tmpmax != (ifreq_flags)tmpmax){
+    if(debug){
+      fprintf(stderr, "Invalid flags \"%s\" for interface \"%s\"\n",
+	      flagstring, if_entry->d_name);
+    }
+    free(flagstring);
+    return 0;
+  }
+  free(flagstring);
+  ifreq_flags flags = (ifreq_flags)tmpmax;
+  /* Reject the loopback device */
+  if(flags & IFF_LOOPBACK){
+    if(debug){
+      fprintf(stderr, "Rejecting loopback interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
+  /* Accept point-to-point devices only if connect_to is specified */
+  if(connect_to != NULL and (flags & IFF_POINTOPOINT)){
+    if(debug){
+      fprintf(stderr, "Accepting point-to-point interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 1;
+  }
+  /* Otherwise, reject non-broadcast-capable devices */
+  if(not (flags & IFF_BROADCAST)){
+    if(debug){
+      fprintf(stderr, "Rejecting non-broadcast interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
+  /* Accept this device */
+  if(debug){
+    fprintf(stderr, "Interface \"%s\" is acceptable\n",
+	    if_entry->d_name);
+  }
+  return 1;
+}
+
 int main(int argc, char *argv[]){
   AvahiSServiceBrowser *sb = NULL;
   int error;
@@ -1027,13 +1126,12 @@
   intmax_t tmpmax;
   char *tmp;
   int exitcode = EXIT_SUCCESS;
-  const char *interface = "eth0";
+  const char *interface = "";
   struct ifreq network;
   int sd = -1;
   bool take_down_interface = false;
   uid_t uid;
   gid_t gid;
-  char *connect_to = NULL;
   char tempdir[] = "/tmp/mandosXXXXXX";
   bool tempdir_created = false;
   AvahiIfIndex if_index = AVAHI_IF_UNSPEC;
@@ -1196,6 +1294,31 @@
   if(not debug){
     avahi_set_log_function(empty_log);
   }
+
+  if(interface[0] == '\0'){
+    struct dirent **direntries;
+    ret = scandir(sys_class_net, &direntries, good_interface,
+		  alphasort);
+    if(ret >= 1){
+      /* Pick the first good interface */
+      interface = strdup(direntries[0]->d_name);
+      if(debug){
+	fprintf(stderr, "Using interface \"%s\"\n", interface);
+      }
+      if(interface == NULL){
+	perror("malloc");
+	free(direntries);
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      free(direntries);
+    } else {
+      free(direntries);
+      fprintf(stderr, "Could not find a network interface\n");
+      exitcode = EXIT_FAILURE;
+      goto end;
+    }
+  }
   
   /* Initialize Avahi early so avahi_simple_poll_quit() can be called
      from the signal handler */
@@ -1272,7 +1395,7 @@
   }
   
   /* If the interface is down, bring it up */
-  if(interface[0] != '\0'){
+  if(strcmp(interface, "none") != 0){
     if_index = (AvahiIfIndex) if_nametoindex(interface);
     if(if_index == 0){
       fprintf(stderr, "No such interface: \"%s\"\n", interface);
@@ -1349,7 +1472,7 @@
       ret = ioctl(sd, SIOCSIFFLAGS, &network);
       if(ret == -1){
 	take_down_interface = false;
-	perror("ioctl SIOCSIFFLAGS");
+	perror("ioctl SIOCSIFFLAGS +IFF_UP");
 	exitcode = EX_OSERR;
 #ifdef __linux__
 	if(restore_loglevel){
@@ -1497,31 +1620,19 @@
     if(quit_now){
       goto end;
     }
-    
-    ret = start_mandos_communication(address, port, if_index, af);
-    if(ret < 0){
-      switch(errno){
-      case ENETUNREACH:
-      case EHOSTDOWN:
-      case EHOSTUNREACH:
-	exitcode = EX_NOHOST;
-	break;
-      case EINVAL:
-	exitcode = EX_USAGE;
-	break;
-      case EIO:
-	exitcode = EX_IOERR;
-	break;
-      case EPROTO:
-	exitcode = EX_PROTOCOL;
-	break;
-      default:
-	exitcode = EX_OSERR;
+
+    while(not quit_now){
+      ret = start_mandos_communication(address, port, if_index, af);
+      if(quit_now or ret == 0){
 	break;
       }
-    } else {
+      sleep(15);
+    };
+
+    if (not quit_now){
       exitcode = EXIT_SUCCESS;
     }
+
     goto end;
   }
   
@@ -1624,7 +1735,7 @@
 	network.ifr_flags &= ~(short)IFF_UP; /* clear flag */
 	ret = ioctl(sd, SIOCSIFFLAGS, &network);
 	if(ret == -1){
-	  perror("ioctl SIOCSIFFLAGS");
+	  perror("ioctl SIOCSIFFLAGS -IFF_UP");
 	}
       }
       ret = (int)TEMP_FAILURE_RETRY(close(sd));

=== modified file 'plugins.d/mandos-client.xml'
--- plugins.d/mandos-client.xml	2009-02-09 02:01:13 +0000
+++ plugins.d/mandos-client.xml	2010-08-10 19:08:24 +0000
@@ -202,8 +202,9 @@
 	<listitem>
 	  <para>
 	    Network interface that will be brought up and scanned for
-	    Mandos servers to connect to.  The default is
-	    <quote><literal>eth0</literal></quote>.
+	    Mandos servers to connect to.  The default is the empty
+	    string, which will automatically choose an appropriate
+	    interface.
 	  </para>
 	  <para>
 	    If the <option>--connect</option> option is used, this
@@ -220,10 +221,11 @@
 	    by this program.
 	  </para>
 	  <para>
-	    <replaceable>NAME</replaceable> can be the empty string;
-	    this will not use any specific interface, and will not
-	    bring up an interface on startup.  This is not
-	    recommended, and only meant for advanced users.
+	    <replaceable>NAME</replaceable> can be the string
+	    <quote><literal>none</literal></quote>; this will not use
+	    any specific interface, and will not bring up an interface
+	    on startup.  This is not recommended, and only meant for
+	    advanced users.
 	  </para>
 	</listitem>
       </varlistentry>

=== modified file 'plugins.d/password-prompt.c'
--- plugins.d/password-prompt.c	2010-03-27 18:39:02 +0000
+++ plugins.d/password-prompt.c	2010-09-07 16:48:58 +0000
@@ -39,11 +39,12 @@
 #include <stdlib.h>		/* EXIT_SUCCESS, EXIT_FAILURE,
 				   getenv() */
 #include <stdio.h>		/* fprintf(), stderr, getline(),
-				   stdin, feof(), perror(), fputc()
+				   stdin, feof(), fputc()
 				*/
 #include <errno.h>		/* errno, EBADF, ENOTTY, EINVAL,
 				   EFAULT, EFBIG, EIO, ENOSPC, EINTR
 				*/
+#include <error.h>		/* error() */
 #include <iso646.h>		/* or, not */
 #include <stdbool.h>		/* bool, false, true */
 #include <string.h> 		/* strlen, rindex */
@@ -70,7 +71,8 @@
 }
 
 int main(int argc, char **argv){
-  ssize_t ret;
+  ssize_t sret;
+  int ret;
   size_t n;
   struct termios t_new, t_old;
   char *buffer = NULL;
@@ -139,7 +141,7 @@
     case ENOMEM:
     default:
       errno = ret;
-      perror("argp_parse");
+      error(0, errno, "argp_parse");
       return EX_OSERR;
     case EINVAL:
       return EX_USAGE;
@@ -155,7 +157,7 @@
   
   if(tcgetattr(STDIN_FILENO, &t_old) != 0){
     int e = errno;
-    perror("tcgetattr");
+    error(0, errno, "tcgetattr");
     switch(e){
     case EBADF:
     case ENOTTY:
@@ -168,17 +170,17 @@
   sigemptyset(&new_action.sa_mask);
   ret = sigaddset(&new_action.sa_mask, SIGINT);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   ret = sigaddset(&new_action.sa_mask, SIGHUP);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   ret = sigaddset(&new_action.sa_mask, SIGTERM);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   /* Need to check if the handler is SIG_IGN before handling:
@@ -187,37 +189,37 @@
   */
   ret = sigaction(SIGINT, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGINT, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
   ret = sigaction(SIGHUP, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGHUP, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
   ret = sigaction(SIGTERM, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGTERM, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
@@ -231,7 +233,7 @@
   t_new.c_lflag &= ~(tcflag_t)ECHO;
   if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_new) != 0){
     int e = errno;
-    perror("tcsetattr-echo");
+    error(0, errno, "tcsetattr-echo");
     switch(e){
     case EBADF:
     case ENOTTY:
@@ -286,11 +288,11 @@
 	}
       }
     }
-    ret = getline(&buffer, &n, stdin);
-    if(ret > 0){
+    sret = getline(&buffer, &n, stdin);
+    if(sret > 0){
       status = EXIT_SUCCESS;
       /* Make n = data size instead of allocated buffer size */
-      n = (size_t)ret;
+      n = (size_t)sret;
       /* Strip final newline */
       if(n > 0 and buffer[n-1] == '\n'){
 	buffer[n-1] = '\0';	/* not strictly necessary */
@@ -298,10 +300,10 @@
       }
       size_t written = 0;
       while(written < n){
-	ret = write(STDOUT_FILENO, buffer + written, n - written);
-	if(ret < 0){
+	sret = write(STDOUT_FILENO, buffer + written, n - written);
+	if(sret < 0){
 	  int e = errno;
-	  perror("write");
+	  error(0, errno, "write");
 	  switch(e){
 	  case EBADF:
 	  case EFAULT:
@@ -318,12 +320,12 @@
 	  }
 	  break;
 	}
-	written += (size_t)ret;
+	written += (size_t)sret;
       }
-      ret = close(STDOUT_FILENO);
-      if(ret == -1){
+      sret = close(STDOUT_FILENO);
+      if(sret == -1){
 	int e = errno;
-	perror("close");
+	error(0, errno, "close");
 	switch(e){
 	case EBADF:
 	  status = EX_OSFILE;
@@ -336,10 +338,10 @@
       }
       break;
     }
-    if(ret < 0){
+    if(sret < 0){
       int e = errno;
       if(errno != EINTR and not feof(stdin)){
-	perror("getline");
+	error(0, errno, "getline");
 	switch(e){
 	case EBADF:
 	  status = EX_UNAVAILABLE;
@@ -352,7 +354,7 @@
 	break;
       }
     }
-    /* if(ret == 0), then the only sensible thing to do is to retry to
+    /* if(sret == 0), then the only sensible thing to do is to retry to
        read from stdin */
     fputc('\n', stderr);
     if(debug and not quit_now){
@@ -368,7 +370,7 @@
     fprintf(stderr, "Restoring terminal attributes\n");
   }
   if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_old) != 0){
-    perror("tcsetattr+echo");
+    error(0, errno, "tcsetattr+echo");
   }
   
   if(quit_now){
@@ -376,7 +378,7 @@
     old_action.sa_handler = SIG_DFL;
     ret = sigaction(signal_received, &old_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     raise(signal_received);
   }

=== added file 'plugins.d/plymouth.c'
--- plugins.d/plymouth.c	1970-01-01 00:00:00 +0000
+++ plugins.d/plymouth.c	2010-09-07 18:48:56 +0000
@@ -0,0 +1,405 @@
+#define _GNU_SOURCE		/* asprintf(), TEMP_FAILURE_RETRY() */
+#include <signal.h>		/* sig_atomic_t, struct sigaction,
+				   sigemptyset(), sigaddset(), SIGINT,
+				   SIGHUP, SIGTERM, sigaction(),
+				   kill(), SIG_IGN */
+#include <stdbool.h>		/* bool, false, true */
+#include <fcntl.h>		/* open(), O_RDONLY */
+#include <iso646.h>		/* and, or, not*/
+#include <sys/types.h>		/* size_t, ssize_t, pid_t, struct dirent,
+				   waitpid() */
+#include <sys/wait.h>		/* waitpid() */
+#include <stddef.h>		/* NULL */
+#include <string.h>		/* strchr(), memcmp() */
+#include <stdio.h>		/* asprintf(), perror(), fopen(), fscanf() */
+#include <unistd.h>		/* close(), readlink(), read(), fork()
+				   setsid(), chdir(), dup2()
+				   STDERR_FILENO, execv(), access() */
+#include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
+#include <dirent.h>		/* scandir(), alphasort() */
+#include <inttypes.h>		/* intmax_t, strtoumax(), SCNuMAX */
+#include <sys/stat.h>		/* struct stat, lstat() */
+#include <sysexits.h>		/* EX_OSERR */
+#include <error.h>		/* error() */
+#include <errno.h>		/* TEMP_FAILURE_RETRY */
+#include <stdarg.h>
+
+sig_atomic_t interrupted_by_signal = 0;
+const char plymouth_pid[] = "/dev/.initramfs/plymouth.pid";
+const char plymouth_path[] = "/bin/plymouth";
+const char plymouthd_path[] = "/sbin/plymouthd";
+const char *plymouthd_default_argv[] = {"/sbin/plymouthd", "--mode=boot",
+					"--attach-to-session",
+					"--pid-file=/dev/.initramfs/plymouth.pid",
+					NULL };
+
+static void termination_handler(__attribute__((unused))int signum){
+  if(interrupted_by_signal){
+    return;
+  }
+  interrupted_by_signal = 1;
+}
+
+/* Create prompt string */
+char *makeprompt(void){
+  int ret = 0;
+  char *prompt;
+  const char *const cryptsource = getenv("cryptsource");
+  const char *const crypttarget = getenv("crypttarget");
+  const char prompt_start[] = "Enter passphrase to unlock the disk";
+  
+  if(cryptsource == NULL){
+    if(crypttarget == NULL){
+      ret = asprintf(&prompt, "%s: ", prompt_start);
+    } else {
+      ret = asprintf(&prompt, "%s (%s): ", prompt_start,
+		     crypttarget);
+    }
+  } else {
+    if(crypttarget == NULL){
+      ret = asprintf(&prompt, "%s %s: ", prompt_start, cryptsource);
+    } else {
+      ret = asprintf(&prompt, "%s %s (%s): ", prompt_start,
+		     cryptsource, crypttarget);
+    }
+  }
+  if(ret == -1){
+    return NULL;
+  }
+  return prompt;
+}
+
+void kill_and_wait(pid_t pid){
+  TEMP_FAILURE_RETRY(kill(pid, SIGTERM));
+  TEMP_FAILURE_RETRY(waitpid(pid, NULL, 0));
+}
+
+bool become_a_daemon(void){
+  int ret = setuid(geteuid());
+  if(ret == -1){
+    error(0, errno, "setuid");
+  }
+    
+  setsid();
+  ret = chdir("/");
+  if(ret == -1){
+    error(0, errno, "chdir");
+    return false;
+  }
+  ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
+  if(ret == -1){
+    error(0, errno, "dup2");
+    return false;
+  }
+  return true;
+}
+
+bool exec_and_wait(pid_t *pid_return, const char *path,
+		   const char **argv, bool interruptable,
+		   bool daemonize){
+  int status;
+  int ret;
+  pid_t pid;
+  pid = fork();
+  if(pid == -1){
+    error(0, errno, "fork");
+    return false;
+  }
+  if(pid == 0){
+    /* Child */
+    if(daemonize){
+      if(not become_a_daemon()){
+	_exit(EX_OSERR);
+      }
+    }
+
+    char **new_argv = NULL;
+    char *tmp;
+    int i = 0;
+    for (; argv[i]!=(char *)NULL; i++){
+      tmp = realloc(new_argv, sizeof(const char *) * ((size_t)i + 1));
+      if (tmp == NULL){
+	error(0, errno, "realloc");
+	free(new_argv);
+	_exit(EXIT_FAILURE);
+      }
+      new_argv = (char **)tmp;
+      new_argv[i] = strdup(argv[i]);
+    }
+    new_argv[i] = (char *) NULL;
+    
+    execv(path, (char *const *)new_argv);
+    error(0, errno, "execv");
+    _exit(EXIT_FAILURE);
+  }
+  if(pid_return != NULL){
+    *pid_return = pid;
+  }
+  do {
+    ret = waitpid(pid, &status, 0);
+  } while(ret == -1 and errno == EINTR
+	  and ((not interrupted_by_signal)
+	       or (not interruptable)));
+  if(interrupted_by_signal and interruptable){
+    return false;
+  }
+  if(ret == -1){
+    error(0, errno, "waitpid");
+    return false;
+  }
+  if(WIFEXITED(status) and WEXITSTATUS(status) == 0){
+    return true;
+  }
+  return false;
+}
+
+int is_plymouth(const struct dirent *proc_entry){
+  int ret;
+  {
+    uintmax_t maxvalue;
+    char *tmp;
+    errno = 0;
+    maxvalue = strtoumax(proc_entry->d_name, &tmp, 10);
+
+    if(errno != 0 or *tmp != '\0' or maxvalue != (uintmax_t)((pid_t)maxvalue)){
+      return 0;
+    }
+  }
+  char exe_target[sizeof(plymouth_path)];
+  char *exe_link;
+  ret = asprintf(&exe_link, "/proc/%s/exe", proc_entry->d_name);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    return 0;
+  }
+
+  struct stat exe_stat;
+  ret = lstat(exe_link, &exe_stat);
+  if(ret == -1){
+    free(exe_link);
+    if(errno != ENOENT){
+      error(0, errno, "lstat");
+    }
+    return 0;
+  }
+  
+  if(not S_ISLNK(exe_stat.st_mode)
+     or exe_stat.st_uid != 0
+     or exe_stat.st_gid != 0){
+    free(exe_link);
+    return 0;
+  }
+
+  ssize_t sret = readlink(exe_link, exe_target, sizeof(exe_target));
+  free(exe_link);
+  if((sret != (ssize_t)sizeof(plymouth_path)-1) or
+      (memcmp(plymouth_path, exe_target,
+	      sizeof(plymouth_path)-1) != 0)){
+    return 0;
+  }
+  return 1;
+}
+
+pid_t get_pid(void){
+  int ret;
+  FILE *pidfile = fopen(plymouth_pid, "r");
+  uintmax_t maxvalue = 0;
+  if(pidfile != NULL){
+    ret = fscanf(pidfile, "%" SCNuMAX, &maxvalue);
+    if(ret != 1){
+      maxvalue = 0;
+    }
+    fclose(pidfile);
+  }
+  if(maxvalue == 0){
+    struct dirent **direntries;
+    ret = scandir("/proc", &direntries, is_plymouth, alphasort);
+    sscanf(direntries[0]->d_name, "%" SCNuMAX, &maxvalue);
+  }
+  pid_t pid;
+  pid = (pid_t)maxvalue;
+  if((uintmax_t)pid == maxvalue){
+    return pid;
+  }
+  
+  return 0;
+}
+
+const char **getargv(pid_t pid){
+  int cl_fd;
+  char *cmdline_filename;
+  ssize_t sret;
+  int ret;
+  
+  ret = asprintf(&cmdline_filename, "/proc/%" PRIuMAX "/cmdline",
+		 (uintmax_t)pid);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    return NULL;
+  }
+  
+  /* Open /proc/<pid>/cmdline  */
+  cl_fd = open(cmdline_filename, O_RDONLY);
+  free(cmdline_filename);
+  if(cl_fd == -1){
+    error(0, errno, "open");
+    return NULL;
+  }
+  
+  size_t cmdline_allocated = 0;
+  size_t cmdline_len = 0;
+  char *cmdline = NULL;
+  char *tmp;
+  const size_t blocksize = 1024;
+  do {
+    /* Allocate more space? */
+    if(cmdline_len + blocksize > cmdline_allocated){
+      tmp = realloc(cmdline, cmdline_allocated + blocksize);
+      if(tmp == NULL){
+	error(0, errno, "realloc");
+	free(cmdline);
+	close(cl_fd);
+	return NULL;
+      }
+      cmdline = tmp;
+      cmdline_allocated += blocksize;
+    }
+    
+    /* Read data */
+    sret = read(cl_fd, cmdline + cmdline_len,
+		cmdline_allocated - cmdline_len);
+    if(sret == -1){
+      error(0, errno, "read");
+      free(cmdline);
+      close(cl_fd);
+      return NULL;
+    }
+    cmdline_len += (size_t)sret;
+  } while(sret != 0);
+  ret = close(cl_fd);
+  if(ret == -1){
+    error(0, errno, "close");
+    free(cmdline);
+    return NULL;
+  }
+  
+  /* we got cmdline and cmdline_len, ignore rest... */
+  const char **argv = NULL;
+  size_t argv_size = 0;
+  for(char *arg = cmdline; arg-cmdline < (ssize_t)cmdline_len;
+      arg = strchr(arg, '\0')+1){
+    tmp = realloc(argv, ((++argv_size)+1)*sizeof(char *));
+    if(tmp == NULL){
+      error(0, errno, "realloc");
+      free(argv);
+      return NULL;
+    }
+    argv = (const char **)tmp;
+    argv[argv_size-1] = arg;
+  }
+  argv[argv_size] = NULL;
+  return argv;
+}
+
+int main(__attribute__((unused))int argc,
+	 __attribute__((unused))char **argv){
+  char *prompt;
+  char *prompt_arg;
+  pid_t plymouth_command_pid;
+  int ret;
+  bool bret;
+
+  /* test -x /bin/plymouth */
+  ret = access(plymouth_path, X_OK);
+  if(ret == -1){
+    exit(EXIT_FAILURE);
+  }
+
+  { /* Add signal handlers */
+    struct sigaction old_action,
+      new_action = { .sa_handler = termination_handler,
+		     .sa_flags = 0 };
+    sigemptyset(&new_action.sa_mask);
+    for(int *sig = (int[]){ SIGINT, SIGHUP, SIGTERM, 0 }; *sig != 0; sig++){
+      ret = sigaddset(&new_action.sa_mask, *sig);
+      if(ret == -1){
+	error(0, errno, "sigaddset");
+	exit(EX_OSERR);
+      }
+      ret = sigaction(*sig, NULL, &old_action);
+      if(ret == -1){
+	error(0, errno, "sigaction");
+	exit(EX_OSERR);
+      }
+      if(old_action.sa_handler != SIG_IGN){
+	ret = sigaction(*sig, &new_action, NULL);
+	if(ret == -1){
+	  error(0, errno, "sigaction");
+	  exit(EX_OSERR);
+	}
+      }
+    }
+  }
+    
+  /* plymouth --ping */
+  bret = exec_and_wait(&plymouth_command_pid, plymouth_path,
+		       (const char *[]){ (const char *)plymouth_path, (const char *)"--ping", (const char *)NULL},
+		       true, false);
+  if(not bret){
+    if(interrupted_by_signal){
+      kill_and_wait(plymouth_command_pid);
+    }
+    exit(EXIT_FAILURE);
+  }
+  
+  prompt = makeprompt();
+  ret = asprintf(&prompt_arg, "--prompt=%s", prompt);
+  free(prompt);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    exit(EXIT_FAILURE);
+  }
+  
+  /* plymouth ask-for-password --prompt="$prompt" */
+  bret = exec_and_wait(&plymouth_command_pid, plymouth_path,
+		       (const char *[]){plymouth_path, "ask-for-password", prompt_arg, NULL},
+		       true, false);
+  free(prompt_arg);
+  if(not bret){
+    if(interrupted_by_signal){
+      kill_and_wait(plymouth_command_pid);
+    } else {
+      exit(EXIT_FAILURE);
+    }
+  }
+  
+  if(bret){
+    exit(EXIT_SUCCESS);
+  }
+  
+  const char **plymouthd_argv = NULL;
+  pid_t pid = get_pid();
+  if(pid == 0){
+    error(0, 0, "plymouthd pid not found");
+  } else {
+    plymouthd_argv = getargv(pid);
+  }
+  if(plymouthd_argv == NULL){
+    plymouthd_argv = plymouthd_default_argv;
+  }
+  
+  bret = exec_and_wait(NULL, plymouth_path,
+  		       (const char *[]){plymouth_path, "quit", NULL}, false, false);
+  if(not bret){
+    exit(EXIT_FAILURE);
+  }
+  bret = exec_and_wait(NULL, plymouthd_path, plymouthd_argv, false, true);
+  if(not bret){
+    exit(EXIT_FAILURE);
+  }
+  exec_and_wait(NULL, plymouth_path,
+  		(const char *[]){ plymouth_path, "show-splash", NULL }, false, false);
+  exit(EXIT_FAILURE);
+}

=== modified file 'plugins.d/splashy.c'
--- plugins.d/splashy.c	2009-10-18 16:09:05 +0000
+++ plugins.d/splashy.c	2010-09-07 16:48:58 +0000
@@ -29,7 +29,7 @@
 				   SIG_IGN, kill(), SIGKILL */
 #include <stddef.h>		/* NULL */
 #include <stdlib.h>		/* getenv() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf() */
 #include <stdlib.h>		/* EXIT_FAILURE, free(),
 				   EXIT_SUCCESS */
 #include <sys/types.h>		/* pid_t, DIR, struct dirent,
@@ -49,6 +49,7 @@
 				   E2BIG, EFAULT, EIO, ETXTBSY,
 				   EISDIR, ELIBBAD, EPERM, EINTR,
 				   ECHILD */
+#include <error.h>		/* error() */
 #include <sys/wait.h>		/* waitpid(), WIFEXITED(),
 				   WEXITSTATUS() */
 #include <sysexits.h>		/* EX_OSERR, EX_OSFILE,
@@ -109,7 +110,7 @@
     proc_dir = opendir("/proc");
     if(proc_dir == NULL){
       int e = errno;
-      perror("opendir");
+      error(0, errno, "opendir");
       switch(e){
       case EACCES:
       case ENOTDIR:
@@ -151,7 +152,7 @@
 	char *exe_link;
 	ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
 	  exitstatus = EX_OSERR;
 	  goto failure;
 	}
@@ -165,7 +166,7 @@
 	    continue;
 	  }
 	  int e = errno;
-	  perror("lstat");
+	  error(0, errno, "lstat");
 	  free(exe_link);
 	  switch(e){
 	  case EACCES:
@@ -213,60 +214,60 @@
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
@@ -283,7 +284,7 @@
     goto failure;
   }
   if(splashy_command_pid == -1){
-    perror("fork");
+    error(0, errno, "fork");
     exitstatus = EX_OSERR;
     goto failure;
   }
@@ -293,7 +294,7 @@
       const char splashy_command[] = "/sbin/splashy_update";
       execl(splashy_command, splashy_command, prompt, (char *)NULL);
       int e = errno;
-      perror("execl");
+      error(0, errno, "execl");
       switch(e){
       case EACCES:
       case ENOENT:
@@ -341,7 +342,7 @@
       goto failure;
     }
     if(ret == -1){
-      perror("waitpid");
+      error(0, errno, "waitpid");
       if(errno == ECHILD){
 	splashy_command_pid = 0;
       }
@@ -379,27 +380,27 @@
 	 the real user ID (_mandos) */
       ret = setuid(geteuid());
       if(ret == -1){
-	perror("setuid");
+	error(0, errno, "setuid");
       }
       
       setsid();
       ret = chdir("/");
       if(ret == -1){
-	perror("chdir");
+	error(0, errno, "chdir");
       }
 /*       if(fork() != 0){ */
 /* 	_exit(EXIT_SUCCESS); */
 /*       } */
       ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace stdout */
       if(ret == -1){
-	perror("dup2");
+	error(0, errno, "dup2");
 	_exit(EX_OSERR);
       }
       
       execl("/sbin/splashy", "/sbin/splashy", "boot", (char *)NULL);
       {
 	int e = errno;
-	perror("execl");
+	error(0, errno, "execl");
 	switch(e){
 	case EACCES:
 	case ENOENT:
@@ -425,13 +426,13 @@
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());

=== modified file 'plugins.d/usplash.c'
--- plugins.d/usplash.c	2010-05-18 16:56:54 +0000
+++ plugins.d/usplash.c	2010-09-07 16:48:58 +0000
@@ -31,19 +31,20 @@
 #include <fcntl.h>		/* open(), O_WRONLY, O_RDONLY */
 #include <iso646.h>		/* and, or, not*/
 #include <errno.h>		/* errno, EINTR */
+#include <error.h>
 #include <sys/types.h>		/* size_t, ssize_t, pid_t, DIR, struct
 				   dirent */
 #include <stddef.h>		/* NULL */
 #include <string.h>		/* strlen(), memcmp() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf()*/
 #include <unistd.h>		/* close(), write(), readlink(),
 				   read(), STDOUT_FILENO, sleep(),
 				   fork(), setuid(), geteuid(),
 				   setsid(), chdir(), dup2(),
 				   STDERR_FILENO, execv() */
 #include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
-				   EXIT_SUCCESS, malloc(), _exit() */
-#include <stdlib.h>		/* getenv() */
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
 #include <dirent.h>		/* opendir(), readdir(), closedir() */
 #include <inttypes.h>		/* intmax_t, strtoimax() */
 #include <sys/stat.h>		/* struct stat, lstat(), S_ISLNK */
@@ -153,7 +154,7 @@
   size_t cmdline_len = 0;
   DIR *proc_dir = opendir("/proc");
   if(proc_dir == NULL){
-    perror("opendir");
+    error(0, errno, "opendir");
     return -1;
   }
   errno = 0;
@@ -181,7 +182,7 @@
       char *exe_link;
       ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
       if(ret == -1){
-	perror("asprintf");
+	error(0, errno, "asprintf");
 	goto fail_find_usplash;
       }
       
@@ -193,7 +194,7 @@
 	  free(exe_link);
 	  continue;
 	}
-	perror("lstat");
+	error(0, errno, "lstat");
 	free(exe_link);
 	goto fail_find_usplash;
       }
@@ -224,13 +225,13 @@
 	ret = asprintf(&cmdline_filename, "/proc/%s/cmdline",
 		       proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
 	  goto fail_find_usplash;
 	}
 	cl_fd = open(cmdline_filename, O_RDONLY);
 	free(cmdline_filename);
 	if(cl_fd == -1){
-	  perror("open");
+	  error(0, errno, "open");
 	  goto fail_find_usplash;
 	}
       }
@@ -242,7 +243,7 @@
 	if(cmdline_len + blocksize > cmdline_allocated){
 	  tmp = realloc(cmdline, cmdline_allocated + blocksize);
 	  if(tmp == NULL){
-	    perror("realloc");
+	    error(0, errno, "realloc");
 	    close(cl_fd);
 	    goto fail_find_usplash;
 	  }
@@ -253,7 +254,7 @@
 	sret = read(cl_fd, cmdline + cmdline_len,
 		    cmdline_allocated - cmdline_len);
 	if(sret == -1){
-	  perror("read");
+	  error(0, errno, "read");
 	  close(cl_fd);
 	  goto fail_find_usplash;
 	}
@@ -261,14 +262,14 @@
       } while(sret != 0);
       ret = close(cl_fd);
       if(ret == -1){
-	perror("close");
+	error(0, errno, "close");
 	goto fail_find_usplash;
       }
     }
     /* Close directory */
     ret = closedir(proc_dir);
     if(ret == -1){
-      perror("closedir");
+      error(0, errno, "closedir");
       goto fail_find_usplash;
     }
     /* Success */
@@ -323,26 +324,26 @@
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -351,7 +352,7 @@
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -360,7 +361,7 @@
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -369,7 +370,7 @@
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -378,7 +379,7 @@
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -387,7 +388,7 @@
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -399,7 +400,7 @@
   /* Write command to FIFO */
   if(not usplash_write(&fifo_fd, "TIMEOUT", "0")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -411,7 +412,7 @@
   
   if(not usplash_write(&fifo_fd, "INPUTQUIET", prompt)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -429,7 +430,7 @@
   outfifo_fd = open("/dev/.initramfs/usplash_outfifo", O_RDONLY);
   if(outfifo_fd == -1){
     if(errno != EINTR){
-      perror("open");
+      error(0, errno, "open");
       status = EX_OSERR;
     }
     goto failure;
@@ -448,7 +449,7 @@
       char *tmp = realloc(buf, buf_allocated + blocksize);
       if(tmp == NULL){
 	if(errno != EINTR){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -460,7 +461,7 @@
 		buf_allocated - buf_len);
     if(sret == -1){
       if(errno != EINTR){
-	perror("read");
+	error(0, errno, "read");
 	status = EX_OSERR;
       }
       TEMP_FAILURE_RETRY(close(outfifo_fd));
@@ -475,7 +476,7 @@
   ret = close(outfifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -488,7 +489,7 @@
   
   if(not usplash_write(&fifo_fd, "TIMEOUT", "15")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -501,7 +502,7 @@
   ret = close(fifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -515,7 +516,7 @@
       sret = write(STDOUT_FILENO, buf + written, buf_len - written);
       if(sret == -1){
 	if(errno != EINTR){
-	  perror("write");
+	  error(0, errno, "write");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -552,7 +553,7 @@
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -561,7 +562,7 @@
   if(outfifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(outfifo_fd));
     if(ret == -1){
-      perror("close");
+      error(0, errno, "close");
     }
   }
   
@@ -575,7 +576,7 @@
 			   (sizeof(char *)
 			    * (size_t)(cmdline_argc + 2)));
       if(tmp == NULL){
-	perror("realloc");
+	error(0, errno, "realloc");
 	free(cmdline_argv);
 	return status;
       }
@@ -602,13 +603,13 @@
        the real user ID (_mandos) */
     ret = setuid(geteuid());
     if(ret == -1){
-      perror("setuid");
+      error(0, errno, "setuid");
     }
     
     setsid();
     ret = chdir("/");
     if(ret == -1){
-      perror("chdir");
+      error(0, errno, "chdir");
       _exit(EX_OSERR);
     }
 /*     if(fork() != 0){ */
@@ -616,13 +617,13 @@
 /*     } */
     ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
     if(ret == -1){
-      perror("dup2");
+      error(0, errno, "dup2");
       _exit(EX_OSERR);
     }
     
     execv(usplash_name, cmdline_argv);
     if(not interrupted_by_signal){
-      perror("execv");
+      error(0, errno, "execv");
     }
     free(cmdline);
     free(cmdline_argv);
@@ -633,7 +634,7 @@
   sleep(2);
   if(not usplash_write(&fifo_fd, "PULSATE", NULL)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
     }
   }
   
@@ -641,7 +642,7 @@
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -652,13 +653,13 @@
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());

